"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

from .scope_config_base import ScopeConfigureBase

class StripToolConfigure(ScopeConfigureBase):
    def __init__(self, params, **kwargs):
        super().__init__(params, **kwargs)

    def assemble_acquisition(self, section=None):
        acquisition = super().assemble_acquisition(section)
        children = acquisition['children']

        children.append({"name": "Sample Mode", "type":"bool", "value": True})

        return acquisition
    
    def assemble_display(self, section=None):
        display = super().assemble_display(section=section, app_section_key="STRIPTOOL", default_autoscale=True)

        return display
    
    def parse(self):
        try:
            acquisition = self.assemble_acquisition(self.params["ACQUISITION"])
        except KeyError:
            acquisition = self.assemble_acquisition()

        try:
            display = self.assemble_display(self.params["DISPLAY"])
        except KeyError:
            display = self.assemble_display()
            
        statistics = self.assemble_statistics()
        # line up in order
        paramcfg = [acquisition, display, statistics]

        return paramcfg





# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2021 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Unit tests for Scope controller

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import os
import sys
import unittest
import configparser
import pvaccess as pva

from pyqtgraph.Qt import QtWidgets
from pyqtgraph.parametertree import Parameter

from c2dataviewer.model import DataSource as DataReceiver
from c2dataviewer.view import Configure
from c2dataviewer.scope import ScopeWindow, WarningDialog
from c2dataviewer.control import ScopeController


os.environ["QT_QPA_PLATFORM"] = "offscreen"


class TestImageDisplay(unittest.TestCase):
    """
    Units tests for the scope controllers.
    """

    def setUp(self):
        """
        Build the environment for each unit test case.
        This method is called before each test.

        :return:
        """

        # Create ImageWindow and get the imageWidget instance
        self.window = ScopeWindow()

        # Build parameter tree
        configure = Configure(configparser.ConfigParser())
        parameters = Parameter.create(
            name="params", type="group", children=configure.parse())
        self.window.parameterPane.setParameters(parameters, showTop=False)

        # Model to be used
        model = DataReceiver()

        # Build GUI elements
        warning = WarningDialog(None)

        self.scope_controller = ScopeController(
            widget=self.window, model=model, parameters=parameters, WARNING=warning)

    def tearDown(self):
        """
        Tear down the environment after each test case.
        This mentod is called after each test.

        :return:
        """
        pass

    def test_update_status(self):
        """
        Test update statistics
        """
        #make sure that update_status runs without errors
        self.scope_controller.update_status()


    def test_get_fdr(self):
        """
        Test if the PV structure is properly parsed and correct fields are
        extracted to be used as a drop down items on the GUI. This are pva.ScalarType
        and arrays of pva.ScalarType.
        """
        # Mock get for the receiver
        def mock_get(*_):

            test_struct = {
                "anInt": pva.INT,
                "aStruct": {
                    "aStruct_aFloat": pva.FLOAT,
                    "aStruct_aDoubleArray": [pva.DOUBLE],
                },
                "aFloatArray": [pva.FLOAT],
                "aStringArray": [pva.STRING],
                "aStructArray": [
                    {
                        "aStructArray_aFloat": pva.FLOAT,
                        "aStructArray_aDoubleArray": [pva.DOUBLE],
                    }
                ],
                "aString": pva.STRING,
                "aDoubleArray": [pva.DOUBLE],
                "aVariant": (),
                "aBooleanArray": [pva.BOOLEAN],
            }
            return pva.PvObject(test_struct)
        self.scope_controller.model.get = mock_get

        fdr, fdr_scalar, fdr_nonnumeric = self.scope_controller.get_fdr()

        self.assertListEqual(fdr, ['aBooleanArray', 'aDoubleArray',
                             'aFloatArray', 'aStruct.aStruct_aDoubleArray'])
        self.assertListEqual(
            fdr_scalar, ['aString', 'aStruct.aStruct_aFloat', 'anInt'])

        self.assertListEqual(fdr_nonnumeric, ['aStringArray'])
        
    def test_buffer_size(self):
        #pass in data
        self.scope_controller.monitor_callback({ 'x' : [10]*100})
        bufval = self.scope_controller.parameters.child("Acquisition").child("Buffer (Samples)").value()
        self.assertEqual(bufval, 100)
        
        #set to object unit
        self.scope_controller.set_buffer_unit('Objects')
        bufval = self.scope_controller.parameters.child("Acquisition").child("Buffer (Objects)").value()
        self.assertEqual(bufval, 1)
        



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import pyqtgraph
import random
from PyQt5 import QtWidgets
from PyQt5 import QtCore
from .pvconfig import PvConfig

def make_color_tuple(c):
    r = int(c[1:3], 16)
    g = int(c[3:5], 16)
    b = int(c[5:7], 16)

    return (r,g,b)

def make_color_hex(c):
    return '#%02x%02x%02x' % (c[0], c[1], c[2])

def randcolor():
    color = "#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)])
    return make_color_tuple(color)

class PvEditDialogController:
    def __init__(self, widget, model, default_proto=None):
        self._win = widget
        self._model = model
        
        self._win.addPvButton.clicked.connect(self._on_addpv_click)
        self._win.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).clicked.connect(self._on_cancel)
        self._win.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(self._on_ok)

        self.callback = None
        self.protocol_list = None
        if default_proto is not None:
            self._win.protocolComboBox.setCurrentText(str(default_proto).lower())

        self._win.colorButton.setColor(randcolor())
            
    def set_completion_callback(self, callback):
        self.callback = callback
            
    def _on_addpv_click(self):
        pv_name = self._win.newPvLineEdit.text()
        proto = self._win.protocolComboBox.currentText()
        color = self._win.colorButton.color(mode='byte')
        self._add_pv(pv_name, color, proto)
        self._win.colorButton.setColor(randcolor())
        
    def _on_cancel(self):
        self._win.close()

    def _on_ok(self):
        pvlist = []
        for row in range(0, self._win.pvTableWidget.rowCount()):
            pvname = self._win.pvTableWidget.item(row, 0).text()
            proto = self._win.pvTableWidget.cellWidget(row, 1).currentText()
            color = make_color_hex(self._win.pvTableWidget.cellWidget(row, 2).color(mode='byte'))
            pvlist.append(PvConfig(pvname, color,proto))
        self.callback(pvlist)
        self._win.close()

    def _get_protocol_list(self):
        if self.protocol_list:
            return self.protocol_list

        self.protocol_list = []
        count = self._win.protocolComboBox.count()
        for i in range(count):
            self.protocol_list.append(self._win.protocolComboBox.itemText(i))
        return self.protocol_list
    
    def _add_pv(self, pvname, color, proto):
        rowcount = self._win.pvTableWidget.rowCount()
        self._win.pvTableWidget.setRowCount(rowcount + 1)

        #pvname edit
        item = QtWidgets.QTableWidgetItem(pvname)
        self._win.pvTableWidget.setItem(rowcount, 0, item)

        #protocol drop-down
        item = QtWidgets.QComboBox()
        item.addItems(self._get_protocol_list())
        item.setCurrentText(str(proto).lower())
        self._win.pvTableWidget.setCellWidget(rowcount, 1, item)

        #color button
        item = pyqtgraph.ColorButton(color=color)        
        self._win.pvTableWidget.setCellWidget(rowcount, 2, item)

        #remove button
        item = QtWidgets.QPushButton('Remove')
        def remove():
            row = self._win.pvTableWidget.indexAt(item.pos()).row();
            self._win.pvTableWidget.removeRow(row)
            
        item.clicked.connect(remove)
        self._win.pvTableWidget.setCellWidget(rowcount, 3, item)
        
    def show(self, pvlist):
        self._set_pvlist(pvlist)
        self._win.exec_()

    def _set_pvlist(self, pvlist):
        self._win.pvTableWidget.clearContents()
        self._win.pvTableWidget.setRowCount(0)
        
        for p in pvlist:
            self._add_pv(p.pvname, make_color_tuple(p.color), p.proto)
        



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2021 UChicago Argonne LLC
 as operator of Argonne National Laboratory

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""

import numpy as np
import pyqtgraph as pg
from .image_definitions import COLOR_MODE_MONO


class ImageProfileWidget(object):
    """
    Part of PVA object viewer utilities, it is responsible to
    display and image profiles.
    """

#########################################
#   Public interface
#########################################

    def __init__(self, canvas_grid_reference):
        super().__init__()

        # Reference to the canvas grid where the image and profiles are displayed.
        self._grid = canvas_grid_reference

        # Internal Numpy array holding image data
        self._data = None
        self._data_color_mode = None

        # Profiles of the data
        self._x_profile = None
        self._y_profile = None

        # Mutex to protect the data variable as set_image_data and plot are
        # called from the different threads
        self._data_mutex = pg.Qt.QtCore.QMutex()

        # Status of the profiles UI
        self._display_widget = False
        self._display_profiles = False
        self._display_rulers = False

        # Image dimensions
        self._nx = 0
        self._ny = 0

        # Save the reference to the image widget
        self._image_widget = self._grid.itemAtPosition(1, 1)

        # Build the X profile graph widget
        self._plot_x_profile = self._grid.itemAtPosition(0, 1)
        self._plot_x_profile.widget().getPlotItem().hideAxis('bottom')
        self._plot_x_profile.widget().getPlotItem().hideAxis('left')
        self._max_x_profile_height = self._plot_x_profile.widget().maximumHeight()
        self._x_axis_min = 0
        self._x_axis_max = 0

        plot_x_size_policy = pg.Qt.QtWidgets.QSizePolicy(pg.Qt.QtWidgets.QSizePolicy.Fixed, pg.Qt.QtWidgets.QSizePolicy.Fixed)
        plot_x_size_policy.setHeightForWidth(self._plot_x_profile.widget().sizePolicy().hasHeightForWidth())
        self._plot_x_profile.widget().setSizePolicy(plot_x_size_policy)

        self._profile_x_curves = {
            'mono'  : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(255,255,255), style=pg.Qt.QtCore.Qt.SolidLine)),
            'red'   : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(255,  0,  0), style=pg.Qt.QtCore.Qt.SolidLine)),
            'green' : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(  0,255,  0), style=pg.Qt.QtCore.Qt.DashLine)),
            'blue'  : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(  0,255,255), style=pg.Qt.QtCore.Qt.DashDotLine)),
        }

        # Build the Y profile graph widget
        self._plot_y_profile = self._grid.itemAtPosition(1, 0)
        self._plot_y_profile.widget().getPlotItem().hideAxis('bottom')
        self._plot_y_profile.widget().getPlotItem().hideAxis('left')
        self._max_y_profile_width = self._plot_y_profile.widget().maximumWidth()
        self._y_axis_min = 0
        self._y_axis_max = 0

        plot_y_size_policy = pg.Qt.QtWidgets.QSizePolicy(pg.Qt.QtWidgets.QSizePolicy.Fixed, pg.Qt.QtWidgets.QSizePolicy.Fixed)
        plot_y_size_policy.setHeightForWidth(self._plot_y_profile.widget().sizePolicy().hasHeightForWidth())
        self._plot_y_profile.widget().setSizePolicy(plot_y_size_policy)

        self._profile_y_curves = {
            'mono'  : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(255,255,255), style=pg.Qt.QtCore.Qt.SolidLine)),
            'red'   : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(255,  0,  0), style=pg.Qt.QtCore.Qt.SolidLine)),
            'green' : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(  0,255,  0), style=pg.Qt.QtCore.Qt.DashLine)),
            'blue'  : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(  0,255,255), style=pg.Qt.QtCore.Qt.DashDotLine)),
        }

        transform = pg.QtGui.QTransform()
        transform.rotate(-90)
        for curve in self._profile_y_curves.values():
            curve.setTransform(transform)

        # Hide profiles by default
        self._plot_x_profile.widget().hide()
        self._plot_y_profile.widget().hide()
        self._grid.removeItem(self._plot_x_profile)
        self._grid.removeItem(self._plot_y_profile)
        self._grid.removeItem(self._image_widget)
        self._grid.addItem(self._image_widget, 0, 0)

    def set_image_data(self, data, color_mode):
        """
        Make an internal copy of the data image.

        :param data: (Numpy arrray) Image of the data as the Numpy array.
        :param color_mode: (enum) Code of the color mode.
        :return: None
        """
        self._data_mutex.lock()
        try:
            shape = data.shape
            if len(shape):
                self._nx = shape[0]
                self._ny = shape[1]
            self._data_color_mode = color_mode
            if not self._display_profiles:
                return
            self._data = np.copy(data)
            self._x_profile, self._y_profile = self._calculate_profiles()
        finally:
            self._data_mutex.unlock()

    def plot(self, display_width, display_height):
        """
        Plot the profiles. This method does nothing if profiles are disabled.

        :param display_width: (int) Image width size on display in pixels.
        :param display_height: (int) Image height size on display in pixels.
        :return: None
        """
        # Return if profiles are not shown
        if not self._display_widget:
            return

        # Setup correct dimensions for the plots
        self._plot_x_profile.widget().setMinimumWidth(int(display_width))
        self._plot_x_profile.widget().setMaximumWidth(int(display_width))
        self._plot_x_profile.widget().setMaximumHeight(self._max_x_profile_height)

        self._plot_y_profile.widget().setMinimumHeight(int(display_height))
        self._plot_y_profile.widget().setMaximumHeight(int(display_height))
        self._plot_x_profile.widget().setMaximumWidth(self._max_y_profile_width)

        if self._x_axis_max:
            self._plot_x_profile.widget().getPlotItem().getAxis('bottom').setRange(self._x_axis_min,self._x_axis_max)
        else:
            self._plot_x_profile.widget().getPlotItem().getAxis('bottom').setRange(0, self._nx)
        if self._y_axis_max:
            self._plot_y_profile.widget().getPlotItem().getAxis('right').setRange(self._y_axis_max,self._y_axis_min)
        else:
            self._plot_y_profile.widget().getPlotItem().getAxis('right').setRange(self._ny, 0)

        if not self._display_profiles:
            self._plot_x_profile.widget().setMaximumHeight(0)
            self._plot_y_profile.widget().setMaximumWidth(0)

        self._grid.update()

        if not self._display_profiles:
            return

        # Calculate profiles
        x_profile, y_profile =  self._x_profile, self._y_profile

        # Draw X profile curves
        self._data_mutex.lock()
        if type(x_profile) is list:
            for mode, curve in self._profile_x_curves.items():
                if mode == 'red':
                    curve.setData(x_profile[0])
                    curve.show()
                elif mode == 'green':
                    curve.setData(x_profile[1])
                    curve.show()
                elif mode == 'blue':
                    curve.setData(x_profile[2])
                    curve.show()
                else:
                    curve.hide()
        else:
            for mode, curve in self._profile_x_curves.items():
                if mode == 'mono':
                    curve.setData(x_profile)
                    curve.show()
                else:
                    curve.hide()

        # Draw Y profile curves
        if type(y_profile) is list:
            for mode, curve in self._profile_y_curves.items():
                if mode == 'red':
                    curve.setData(y_profile[0])
                    curve.show()
                elif mode == 'green':
                    curve.setData(y_profile[1])
                    curve.show()
                elif mode == 'blue':
                    curve.setData(y_profile[2])
                    curve.show()
                else:
                    curve.hide()
        else:
            for mode, curve in self._profile_y_curves.items():
                if mode == 'mono':
                    curve.setData(y_profile)
                    curve.show()
                else:
                    curve.hide()
        self._data_mutex.unlock()

    def setXAxisRange(self, xmin=0, xmax=0):
        """
        Set range for x axis.

        :param xmin: (float) x axis min value
        :param xmax: (float) x axis max value
        :return:
        """
        self._x_axis_min = xmin
        self._x_axis_max = xmax

    def setYAxisRange(self, ymin=0, ymax=0):
        """
        Set range for y axis.

        :param ymin: (float) y axis min value
        :param ymax: (float) y axis max value
        :return:
        """
        self._y_axis_min = ymin
        self._y_axis_max = ymax

    def showProfiles(self, displayProfiles=None):
        """
        Display profiles. This method should be called from the UI thread.

        :param displayProfiles: (bool) True to show the profiles, False otherwise.
        :return:
        """
        if displayProfiles is not None:
            self._display_profiles = displayProfiles

        if not self._display_profiles:
            for curve in self._profile_x_curves.values():
                curve.hide()
            for curve in self._profile_y_curves.values():
                curve.hide()

    def showRulers(self, displayRulers=None):
        """
        Display rulers. This method should be called from the UI thread.

        :param displayRulers: (bool) True to show the rulers, False otherwise.
        :return:
        """
        if displayRulers is not None:
            self._display_rulers = displayRulers
        if self._display_rulers:
            self._plot_x_profile.widget().getPlotItem().showAxis('bottom')
            self._plot_y_profile.widget().getPlotItem().showAxis('right')
        else:
            self._plot_x_profile.widget().getPlotItem().hideAxis('bottom')
            self._plot_y_profile.widget().getPlotItem().hideAxis('right')

    def show(self, displayWidget):
        """
        Hide or show the widget. This method should be called from the UI thread.

        :param displayWidget: (bool) True to show the widget, False otherwise.
        :return:
        """
        if displayWidget:
            self._grid.removeItem(self._image_widget)
            self._grid.addItem(self._image_widget, 1, 1)
            self._grid.addItem(self._plot_x_profile, 0, 1, 1, 1, pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignBottom)
            self._grid.addItem(self._plot_y_profile, 1, 0, 1, 1, pg.Qt.QtCore.Qt.AlignRight | pg.Qt.QtCore.Qt.AlignTop)
            self._plot_x_profile.widget().show()
            self._plot_y_profile.widget().show()
            self._display_widget = True
        else:
            self._grid.removeItem(self._image_widget)
            self._grid.removeItem(self._plot_x_profile)
            self._grid.removeItem(self._plot_y_profile)
            self._grid.addItem(self._image_widget, 1, 1)
            self._plot_x_profile.widget().hide()
            self._plot_y_profile.widget().hide()
            self._display_widget = False

        self._grid.update()

#########################################
#   Internal methods
#########################################

    def _calculate_profiles(self):
        """
        Calculate the profiles.

        :return: (numpy array, numpy array) X and Y profiles of the image.
        """
        # Calculate the profiles
        x_profile = y_profile = None
        if self._data_color_mode == COLOR_MODE_MONO:
            x_profile = self._data.mean(axis=1)
            y_profile = self._data.mean(axis=0)
        else:
            x_profile = []
            y_profile = []
            for i in range(3):
                x_profile.append(self._data[:,:,i].mean(axis=1))
                y_profile.append(self._data[:,:,i].mean(axis=0))

        return x_profile, y_profile



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities for image display

@author: Guobao Shen <gshen@anl.gov>
"""
from datetime import datetime
import numpy as np
import pyqtgraph
from pyqtgraph.Qt import QtCore
import enum
import logging
import time

class PlotChannel:
    """
    Channel data to plot
    """
    
    def __init__(self, pvname, color):
        """
        pvname: (str) PV data name.  Either is the PV name or PV.field1.field2 format.  
        """
        self.color = color
        self.pvname = pvname
        self.dc_offset = 0
        self.axis_location = 'left'
        

class TriggerType(enum.Enum):
    ON_CHANGE = 'onchange'
    GT_THRESHOLD = 'gtthreshold'
    LT_THRESHOLD = 'ltthreshold'

    def __str__(self):
        return self.value
    
class Trigger:
    def __init__(self, parent):
        self.parent = parent
        self.trig_data = {}
        
        # Is trigger mode enabled
        self.trigger_mode = False

        # True if in trigger mode and trigger was received (trigger PV monitor fired).
        self.is_triggered_ = False

        # Counts trigger PV monitor fired callback (count monitor updates)
        self.trigger_count = 0

        #trigger position in trig_data
        self.trigger_index = 0
        #index in trig_data where to plot data
        self.display_start_index = 0

        self.trigger_data_done = True
        
        # double sec past epoch timestamp from the trig pv
        self.trigger_timestamp = 0.0

        self.trigger_level = 0.0
        class TriggerSignal(QtCore.QObject):
            my_signal = QtCore.pyqtSignal()

            def emit(self):
                self.my_signal.emit()
                
            def __init__(self):
                QtCore.QObject.__init__(self)

        self.plot_trigger_signal_emitter = TriggerSignal()
        
        self.trigger_type = None
        self.trigger_time_field = 'timeStamp'
        self.data_time_field = None
        self.is_triggered_func = None

        self.trigger_value = None
        self.missed_triggers = 0
        self.missed_adjust_buffer_size = 0
        
    def __trigger_on_change(self, val):
        return True

    def __trigger_gt_threshold(self, val):
        return val > self.trigger_level
    
    def __trigger_lt_threshold(self, val):
        return val < self.trigger_level
    
    def set_trigger_type(self, val):
        func_table = { TriggerType.ON_CHANGE: self.__trigger_on_change,
                       TriggerType.GT_THRESHOLD: self.__trigger_gt_threshold,
                       TriggerType.LT_THRESHOLD: self.__trigger_lt_threshold }
        self.trigger_type = TriggerType(val)
        self.is_triggered_func = func_table[self.trigger_type]
        
    def set_trigger_mode(self, flag):
        self.trigger_mode = flag

    def is_triggered(self):
        return self.trigger_mode and self.is_triggered_
    
    def data_callback(self, data):
         # Count triggers and ignore first one which is initial connection callback and not actual value change.
        self.trigger_count += 1
        if self.trigger_count <= 1:
            return

        self.trigger_value = data['value']

        if not self.parent.plotting_started:
            return

        if not self.trigger_data_done:
            return

        if not self.is_triggered_func(data['value']):
            return

        self.is_triggered_ = True
        self.trigger_data_done = False

        ts = data[self.trigger_time_field]
        self.trigger_timestamp = ts['secondsPastEpoch'] + 1e-9*ts['nanoseconds']

    def __is_trigger_in_array(self, time_array):
        """
        Calculate if the trigger timestap is in the timestamp array.

        :param time_array: (numpy array) Time (SORTED!!!) array to be checked.
        :return: (Tuple -> (bool, integer)) True or False if timestamp is in the array.
                                            If True, second element hold index of the element, otherwise is None.
        """

        if self.trigger_timestamp < time_array[0]:
            return -1

        if self.trigger_timestamp > time_array[-1]:
            return -2
        
        idx = np.searchsorted(time_array, self.trigger_timestamp)

        return idx

    def status(self):
        if not self.trigger_mode:
            return 'Off'
        
        if self.missed_triggers > 0:
            return 'Trig off by %.1f s (Set buf=%.1e)' % (self.missed_time, self.missed_adjust_buffer_size)
        elif self.is_triggered_:
            return 'Collecting Data'
        else:
            return 'Waiting for trigger'

    def __trig_max_length(self):
        return int(1.5 * self.parent.max_length)
    
    def add_to_trig_data(self, k, v):
        # In trigger mode the minimum we must save is the whole last array (+ part of the old ones if max_length > vector_len)
        # We must store all the arrayes as the "Time" field determinate how much data do we really need, and the "Time" field can arrive last.
        required_data = self.__trig_max_length()
        self.trig_data[k] = np.append(self.trig_data.get(k, []), v)[-required_data:]

    def draw_data(self):
        """
        If collected max_length/2 samples after the trigger mark, call signal to draw data
        """
        max_length = self.parent.max_length
        
        if self.trigger_mode and self.is_triggered_:
            samples_after_trig = int(max_length / 2)

            # Check if we have trigger timestamp in buffer
            time_data = self.trig_data[self.data_time_field]
            idx = self.__is_trigger_in_array(time_data)
            if idx >= 0:
                self.missed_triggers = 0                
                self.trigger_index = idx
                samples_after_trig_cnt = time_data.size - idx
                # Check if we have the requested number of samples, to trigger the plotting
                if samples_after_trig_cnt >= samples_after_trig:
                    self.display_start_index = max(idx - samples_after_trig, 0)
                    endindex = idx + samples_after_trig
                    for k, v in self.trig_data.items():
                        if type(v) != np.ndarray:
                            continue

                        if len(self.trig_data[k]) < time_data.size:
                            continue
                        
                        # `idx` holds the index in the stored waveforms where the trigger happened.
                        # Goal is to get the slice of the waveform so the sample when the trigger happened is in the middle.
                        # samples_after_trig determinate how much samples do we want to show after the trigger (idx + samples_after_trig),
                        # while max_length holds a maximum number of samples to show, for this reason we need `(self.max_length-self.samples_after_trig`
                        # samples before trigger/idx.
                        self.parent.data[k] = self.trig_data[k][self.display_start_index : endindex]
                    self.plot_trigger_signal_emitter.emit()
                else:
                    logging.getLogger().debug('Needs to acquire %i more samples before plotting' % (samples_after_trig - samples_after_trig_cnt))
            elif idx == -1:
                if len(time_data) >= self.__trig_max_length():
                    datatimediff = time_data[-1] - time_data[0]
                    timediff = time_data[0] - self.trigger_timestamp
                    increasesize = len(time_data) * timediff / datatimediff
                    self.missed_triggers += 1
                    self.missed_adjust_buffer_size = increasesize + self.parent.max_length
                    self.missed_time = timediff
                    self.is_triggered_ = False
                    self.trigger_data_done = True
                else:
                    self.missed_triggers = 0
            else:
                logging.getLogger().debug("Data is %f seconds behind trigger time" % (self.trigger_timestamp - time_data[-1]))

    def display_trigger_index(self):
        """
        Returns the position of the trigger in the display buffer (i.e self.parent.data)
        """
        return self.trigger_index - self.display_start_index
    
    def finish_drawing(self):
        # Trigger curves were drawn - Freeze the scope
        if self.trigger_mode and self.is_triggered_:
            self.is_triggered_ = False
            self.trigger_data_done = True

    def reset(self):
        self.trigger_count = 0
        self.is_triggered_ = False
        self.trigger_data_done = True
        self.trigger_value = None
        self.missed_triggers = 0
        
    def connect_to_callback(self):
        if self.trigger_mode:
            self.plot_trigger_signal_emitter.my_signal.connect(self.parent.update_drawing)

    def disconnect_to_callback(self):
        try:
            self.plot_trigger_signal_emitter.my_signal.disconnect()
        except Exception:
            pass

class MouseOver:
        
    def __init__(self, widget):
        self.display_fields = []
        self.widget = widget
        self.textbox = None
        self.vline = None
        self.hline = None
        self.enabled = False
        self.display_location = None
        self.mouse_index = None
        
    def setup_plot(self):
        assert(self.widget.plot)
        self.textbox = pyqtgraph.LabelItem(color='w')
        self.textbox.setAttr('size', '12pt')
        self.vline = pyqtgraph.InfiniteLine(angle=90, movable=False)
        self.hline = pyqtgraph.InfiniteLine(angle=0, movable=False)

        self.enable(self.enabled)
        
    def enable(self, flag):
        self.enabled = flag
        if flag:
            self.widget.plot.addItem(self.vline, ignoreBounds=True)
            self.widget.plot.addItem(self.hline, ignoreBounds=True)
            self.textbox.setParentItem(self.widget.plot)
            self._apply_display_location()
            self.textbox.show()
            self.vline.show()
            self.hline.show()
        else:
            self.widget.plot.removeItem(self.vline)
            self.widget.plot.removeItem(self.hline)
            self.textbox.setParentItem(None)

            self.textbox.hide()
            self.vline.hide()
            self.hline.hide()
            
    def set_display_fields(self, fields):
        self.display_fields = fields
        
    def _apply_display_location(self):
        if self.enabled:
            self.textbox.anchor(parentPos=(1,0.95), itemPos=(1,1))
            if self.display_location == 'top-right':
                 self.textbox.anchor(parentPos=(1, 0.00), itemPos=(1,0))
            elif self.display_location == 'bottom-right':
                 self.textbox.anchor(parentPos=(1, 0.95), itemPos=(1,1))
            elif self.display_location == 'bottom-left':
                 self.textbox.anchor(parentPos=(0.05, 0.95), itemPos=(0,1))
        
    def set_display_location(self, loc):
        self.display_location = loc
        self._apply_display_location()

    def update_textbox(self):
        if self.mouse_index is None or not self.enabled:
            return

        index = self.mouse_index
        channel_data = {}
        nsamples = 0

        for ch in self.widget.channels:
            if ch.pvname == 'None':
                continue
            if ch.pvname not in self.widget.data:
                continue
            
            channel_data[ch.pvname] = [self.widget.data.get(ch.pvname), ch.color]
            nsamples = max(nsamples, len(channel_data[ch.pvname][0]))

        for f in self.display_fields:
            if f in channel_data:
                continue

            channel_data[f] = [self.widget.data.get(f), '#FFFFFF']
            try:
                nsamples = max(nsamples, len(channel_data[f][0]))
            except:
                pass
            
        if len(channel_data) == 0:
            return

        if index >= 0 and index < nsamples:
            text = []
            xaxis = self.widget.current_xaxes
            if xaxis == 'None':
                text.append(f"x={index}")
            else:
                text.append(f"{xaxis}={self.widget.data.get(xaxis)}")
                
            for k,v in channel_data.items():
                data = v[0]
                color = v[1]
                data_value = data[index] if data is not None and index <len(data) else 'N/A'
                text.append(f"<span style='color: {color}'>{k}={data_value}</span>")
                
            text = "<br>".join(text)
            self.textbox.setText(text)
        
    def on_mouse_move_event(self, event):
        if not self.enabled:
            return
        pos = event.pos()
                
        if not self.widget.plot.sceneBoundingRect().contains(pos):
            return

        mousePoint = self.widget.plot.vb.mapSceneToView(pos)
        self.mouse_index = int(mousePoint.x())

        self.update_textbox()
        
        self.vline.setPos(mousePoint.x())
        self.hline.setPos(mousePoint.y())
            
class PlotWidget(pyqtgraph.GraphicsLayoutWidget):
    """
    Main class for plotting multiple sets of data.  Key functions:
    - setup_plot - sets up the channels to plot
    - data_process - pass in data to plot
    - update_drawing - plots data
    """
    def __init__(self, parent=None, **kargs):
        pyqtgraph.GraphicsLayoutWidget.__init__(self, parent=parent)
        self.setParent(parent)

        self.param_changed = False
        self.model = None
        self.channels = []
        self.auto_scale = False
        self.first_data = True
        self.plotting_started = False
        
        self.sampling_mode = False        
        self.sample_data = {}
        
        self.mutex = pyqtgraph.QtCore.QMutex()

        self.max_length = None
        self.data_size = 0
        self.new_buffer = True
        self.data = {}
        self.first_run = True
        self.new_plot = True
        
        # last id number of array received
        self.lastArrayId = None
        # count of lost arrays
        self.lostArrays = 0
        # count of array received
        self.arraysReceived = 0

        # EPICS7 field name for horizontal axis
        self.current_xaxes = "None"
        # EPICS7 field name for Array ID
        self.current_arrayid = "None"

        self.major_ticks = 0
        self.minor_ticks = 0
        
        self.fps = 0
        self.lastTime = time.time()

        self._max = None
        self._min = None

        self.bins = 100
        self.fft = False
        self.fft_filter = None
        self.psd = False
        self.diff = False
        self.xy = False
        self.histogram = False
        self.average = 1

        self.__fft_vgain = {
            'none' : 1,
            'hamming' : 1.853,
        }

        self.axis = []
        self.views = []
        self.curves = []

        
        self.trigger = Trigger(self)
        self.mouse_over = MouseOver(self)
        
        # Setup plot variables
        self.single_axis = True
        self.plot = None
        self.setup_plot([])

        self.is_freeze = False

    def trigger_mode(self):
        return self.trigger.trigger_mode

    def set_trigger_mode(self, flag):
        self.trigger.set_trigger_mode(flag)
            
    def set_model(self, model):
        """

        :param model:
        :return:
        """
        self.model = model


    def enable_sampling_mode(self, val):
        """
        Set sampling mode.  If true, plot only the latest value at each refresh interval.
        Set this mode if plotting multiple data sources that have different data rates.
        """
        if self.sampling_mode != val:
            self.sampling_mode = val
            self.data.clear()
            self.setup_plot()
        
    def set_arrayid(self, value):
        """
        Set current field name for array id

        :param value:
        :return:
        """
        if value != self.current_arrayid:
            self.current_arrayid = value

    def set_xaxes(self, value):
        """
        Set current field name for x axes

        :param value:
        :return:
        """
        self.current_xaxes = value
        self.new_plot = True
            
        if self.current_xaxes == "None":
            self.plot.setLabel('bottom', '')
        else:
            self.plot.setLabel('bottom', self.current_xaxes)

    def set_enable_mouseover(self, value):
        self.mouse_over.enable(value)

    def set_mouseover_fields(self, fields):
        self.mouse_over.set_display_fields(fields)

    def set_mouseover_display_location(self, loc):
        self.mouse_over.set_display_location(loc)
        
    def mouseMoveEvent(self, event):
        pyqtgraph.GraphicsLayoutWidget.mouseMoveEvent(self, event)
        self.mouse_over.on_mouse_move_event(event)
        
    def set_major_ticks(self, value):
        self.major_ticks = value

    def set_minor_ticks(self, value):
        self.minor_ticks = value

    def _setup_ticks(self):
        if self.major_ticks <= 0 or self.current_xaxes != 'None':
            self.plot.getAxis('bottom').setTicks(None)
            return
        
        minVal = 0
        maxVal = self.max_length

        major_ticks = [ (i, str(i)) for i in range(0, maxVal + 1, int(self.major_ticks)) ]

        minor_ticks = None
        if self.minor_ticks > 0:
            tickMod = self.major_ticks / self.minor_ticks
            minor_ticks = [ (i, str(i) if (i/self.minor_ticks % tickMod) > 0 else '') for i in range(0, maxVal + 1, int(self.minor_ticks)) ]

        ticks = [major_ticks]
        if minor_ticks:
            ticks.append(minor_ticks)
            
        self.plot.getAxis('bottom').setTicks(ticks)
                                 
            
    def setup_plot(self, channels=None, single_axis=None):
        """
        Setup the channels to plot

        :param channels: list of PlotChannel objects
        :param single_axis: flag to share single axis, or have a axis for each figure. FFT and PSD support only single axis
        :return:
        """
        if single_axis is None:
            single_axis = self.single_axis

        # FFT and PSD support only single axis, PyQTGraph currently doesn't support log scale on the multiple axis setup
        if not single_axis and (self.fft or self.psd):
            single_axis = True
            # Could Be replaced with logging if added in the future
            print(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}: FFT or PSD selected in multi axis mode, which is not supported. Changed to one axis mode automatically.")

        # Delete existing plots
        self.delete_plots()

        # Set new list of signals
        if channels:
            self.channels = channels

        # Create plot item
        self.plot = pyqtgraph.PlotItem()
        self.plot.showGrid(x=True, y=True)
        if self.fft or self.psd:
            self.plot.setLogMode(x=True, y=True)

        # Generate plot items
        self.single_axis = single_axis
        if single_axis:
            self.setup_plot_single_axis()
        else:
            self.setup_plot_multi_axis()

        self.trigMarker = self.plot.plot()

        # Update plots
        self.plot.vb.sigResized.connect(self.update_views)
        self.update_views()

        #apply settings to new plot
        self.set_autoscale(self.auto_scale)
        self.set_xaxes(self.current_xaxes)
        
        self.new_plot = True
        self.data_size = 0
        self.mouse_over.setup_plot()
        
    def setup_plot_single_axis(self):
        """
        Setup items for single Y axis plot.

        :names: (list -> strings) List of channels to bi displayed. "None" should be used to ignore that channel.
        :return: (None)
        """
        if self.channels:
            self.plot.addLegend()
        for ch in self.channels:
            if ch.pvname == "None":
                continue
            curve = self.plot.plot(pen=ch.color, name=ch.pvname)
            curve.plotdata_ave = None
            self.curves.append(curve)

        self.ci.addItem(self.plot, row=2)

    def setup_plot_multi_axis(self):
        """
        Setup items for multiple Y axis plot.

        :names: (list -> strings) List of channels to bi displayed. "None" should be used to ignore that channel.
        :return: (None)
        """
        # Workaround until pyqtgraph issue with multiple axis is resolved
        # (see https://git.aps.anl.gov/C2/conda/data-viewer/-/issues/48)
        # Pick up proper viewbox geometry from parent
        self.ci.addItem(self.plot, row=2)

        # We do not use default axis
        self.plot.hideAxis('left')

        # Create axis and views
        left_axis = []
        right_axis = []
        for i, ch in enumerate(self.channels):
            if ch.pvname == "None":
                continue

            axis = pyqtgraph.AxisItem(ch.axis_location)
            axis.setLabel(f"Channel {i+1} [{ch.pvname}]", color=ch.color)
            if ch.axis_location == "left":
                left_axis.append(axis)
            else:
                right_axis.append(axis)
            self.axis.append(axis)
            # First view we take from plotItem
            if len(self.axis) == 1:
                self.views.append(self.plot.vb)
            else:
                self.views.append(pyqtgraph.ViewBox())

        # Add axis and plot to GraphicsLayout (self.ci)
        # Order how they are added is important, while it is also important
        # we hold correct order (the same as in the "names" variable) in self.axis
        # variable as other code depends on this.
        for a in left_axis:
            self.ci.addItem(a, row=2)
        self.ci.addItem(self.plot, row=2)
        for a in right_axis:
            self.ci.addItem(a, row=2)

        # Add Viewboxes to the layout
        for vb in self.views:
            if vb != self.plot.vb:
                self.ci.scene().addItem(vb)

        # Link axis to view boxes
        for i, view in enumerate(self.views):
            self.axis[i].linkToView(view)
            if view != self.plot.vb:
                view.setXLink(self.plot.vb)

        # Make curves
        count = 0

        for i, ch in enumerate(self.channels):
            if ch.pvname == 'None':
                continue
            
            curve = pyqtgraph.PlotCurveItem(pen=ch.color)
            curve.plotdata_ave = None
            left_curve = []
            right_curve = []
            if ch.axis_location == "left":
                left_curve.append(curve)
            else:
                right_curve.append(curve)
            self.curves.extend(left_curve)
            self.curves.extend(right_curve)
            self.views[count].addItem(curve)
            count += 1

    def delete_plots(self):
        """
        Delete all plots

        :return:
        """
        if self.single_axis:
            for curve in self.curves:
                self.plot.removeItem(curve)
        else:
            # Delete old plots if exists
            for i, view in enumerate(self.views):
                view.removeItem(self.curves[i])

            for a in self.axis:
                self.ci.removeItem(a)

        # Remove plotItem
        if self.plot is not None:
            self.ci.removeItem(self.plot)

        # Remove references to all chart items
        self.views = []
        self.axis = []
        self.curves = []

    def update_views(self):
        """
        Update the view so they scale properly.

        :return:
        """
        for view in self.views:
            if view == self.plot.vb:
                continue
            view.setGeometry(self.plot.vb.sceneBoundingRect())

    def set_autoscale(self, flag):
        """
        Enable or disable the autoscale.

        :param flag: (bool) True to enable the autoscale, False otherwise.
        :return:
        """
        self.auto_scale = flag

        if self.auto_scale:
            self.plot.enableAutoRange()
            for view in self.views:
                view.enableAutoRange()
        else:
            self.plot.disableAutoRange()
            for view in self.views:
                view.disableAutoRange()
                    


    def do_autoscale(self):
        """
        Auto-scale x/y range of the current plot
        """
        #Use autoRange instead enableAutoRange when autoscale not set so that 
        #it isn't persistant. This way we can use this function for one time 
        #auto-scale and for persistant autoscales
        if self.auto_scale:
            self.plot.enableAutoRange()
            for view in self.views:
                view.enableAutoRange()
        else:
            self.plot.autoRange()
            for view in self.views:
                view.autoRange()

                
    def reset_xrange(self):
        """
        Set x range of the current plot to maximum buffer size
        Not supported cases where the x values are not the buffer samples
        """

        #
        # Cases where x-values are not buffer samples
        #
        if self.current_xaxes != None or self.fft or self.psd or self.histogram:
            return
        
        xmin = 0;
        xmax = self.max_length
        self.plot.setXRange(xmin, xmax)
        for view in self.views:
            view.setXRange(xmin, xmax)
            
    def wait(self):
        """

        :return:
        """
        self.mutex.lock()

    def signal(self):
        """

        :return:
        """
        self.mutex.unlock()

    def update_buffer(self, value):
        """

        :param value:
        :return:
        """
        self.max_length = value
        self.new_buffer = True
        
    def set_range(self, **kwargs):
        """
        Set data range for data filtering.
        Any data out of range will be filtered out, and not plotted.

        :param kwargs:
        :return:
        """
        self._max = kwargs.get("max", None)
        self._min = kwargs.get("min", None)

    def filter(self, data, pair=None):
        """
        Filter data out of range

        :param data:
        :param pair:
        :return:
        """
        rd, rt = data, pair
        if self._max is not None:
            mask = rd <= self._max
            rd = rd[mask]
            if pair is not None:
                rt = rt[mask]
        if self._min is not None:
            mask = rd >= self._min
            rd = rd[mask]
            if pair is not None:
                rt = rt[mask]
        if pair is not None:
            return rd, rt
        return rd

    def set_display_mode(self, value):
        """

        :param value:
        :return:
        """
        if value == 'normal':
            self.fft = False
            self.psd = False
            self.diff = False
            self.xy = False
        elif value == 'fft':
            self.fft = True
            self.psd = False
            self.diff = False
            self.xy = False
        elif value == 'psd':
            self.fft = False
            self.psd = True
            self.diff = False
            self.xy = False
        elif value == 'diff':
            self.fft = False
            self.psd = False
            self.diff = True
            self.xy = False
        elif value == 'xy':
            self.fft = False
            self.psd = False
            self.diff = False
            self.xy = True
        self.plot.autoScale = True

        # Mode changed, iterate over the curves and set moving average to None
        for curve in self.curves:
            curve.plotdata_ave = None

    def set_fft_filter(self, value):
        """
        Set fft filter to use. Valid values are None / "none" and "hamming".

        :param value: (None or string) Filter to be used on the data, before the fft.
        :return:
        """
        if value is None:
            self.fft_filter = value
        else:
            self.fft_filter = value.lower()

    def set_average(self, value):
        """
        Set average number

        :param value:
        :return:
        """
        self.average = value

    def set_histogram(self, flag):
        """
        Set flag to enable/disable histogram plotting

        :param flag:
        :return:
        """
        if flag != self.histogram:
            self.new_plot = True

        self.histogram = flag

    def set_binning(self, value):
        """
        Set number for binning

        :param value:
        :return:
        """
        self.bins = value
   
    def clear_sample_data(self, pvname):
        """
        Clear sample data for PV, usually when the connection is lost.  
        """
        if pvname in self.sample_data:
            self.sample_data[pvname] = 0

    def data_process(self, data_generator):
        """
        Process raw data off the wire

        :param data_generator: generator that yields pvaccess data from the wire as key/value pairs. 
        :return:
        """
        # Increment array count
        self.arraysReceived += 1

        # Check if scope is frozen
        if self.is_freeze:
            return

        self.wait()
        new_size = self.data_size
        vector_len = 0

        got_data = False
        
        for k, v in data_generator():
            if k == self.current_arrayid:
                if self.lastArrayId is not None:
                    if v - self.lastArrayId > 1:
                        self.lostArrays += 1
                self.lastArrayId = v
                if self.data_size == 0:
                    new_size += 4

            # If pvaccess modules was build without numpy, normal Python list is returned. Make conversion here.
            if type(v) is list:
                v = np.array(v)

            if np.isscalar(v):
                v = np.array([v])
                
            if type(v) is not np.ndarray:
                continue

            # At this point in program we found an nd array that may or may not have data. nd array is
            # not an EPICS7 NDArray, but a numpy ndarray.
            # If no data in a pv field, we just skip this whole loop iteration.
            if len(v) == 0:
                continue

            if not np.isscalar(v[0]):
                continue

            if self.data_size == 0:
                type_size = v[0].dtype.itemsize
                new_size += type_size * len(v)
                
            got_data = True

            if self.sampling_mode:
                self.sample_data[k] = v[-1]
            elif self.trigger.trigger_mode:
                self.trigger.add_to_trig_data(k, v)
            else:
                # We are in free running mode. Only last max_length of the data can be stored.
                arr = np.append(self.data.get(k, []), v)
                self.data[k] = arr[-self.max_length:]
        if got_data:
            self.trigger.draw_data()
                
        self.signal()
        if self.data_size == 0:
            self.data_size = new_size
            
        if self.first_data:
            self.first_data = False

    def calculate_ftt(self, data, sample_period, mode, filter=None):
        """
        Calculate fft on the input array.

        :param data: (Numpy array) Data to be transformed.
        :param sample_period: (int) Sampling period of the data.
        :param mode: (string) This could be either "fft" or "psd".
        :param filter: (None or string) Filter to be applied on the data before transformation.

        :return: (Numpy array, Numpy array) xf and yf transformations of the data.
        """
        xf = yf = None
        filter = filter or "none"

        # Perform transformation for X axis
        data_len = len(data)
        xf = np.fft.rfftfreq(data_len, d=sample_period)

        # Calculate window
        if filter == "none":
            pass
        elif filter == "hamming":
            window = np.hamming(data_len)
            data = data * window
        else:
            raise ValueError(f"{str(filter)} is not valid FFT filter type.")

        # Perform transformation for Y axis
        yf_raw = np.abs(np.fft.rfft(data))

        # Apply normalisation filter
        try:
            vertical_gain = self.__fft_vgain[filter]
        except KeyError:
            raise ValueError(f"{str(filter)} is not valid FFT filter type.")

        if mode.lower() == "fft":
            yf = 2. * vertical_gain * yf_raw / data_len
            # DC bin has different scale factor
            yf[0] = yf[0] / 2.
        elif mode.lower() == "psd":
            # Sample period in sec. Sample rate is 1/sample rate. Hz/bin=srage/nf
            sample_rate = 1. / sample_period
            yf = 2. * vertical_gain * vertical_gain * yf_raw * yf_raw / (sample_rate * sample_rate)
            # DC bin has different scale factor
            yf[0] = yf[0] / 4.
        else:
            raise ValueError(f"{str(mode)} is not valid mode for FFT.")

        return xf, yf

    def exponential_moving_average(self, data, data_ave):
        """
        Calculate exponential moving average on the data.

        :param data: (ndarray) Latest array data.
        :param data_ave: (ndarray) Data average until now.

        :return: (ndarray) Exponential moving average.
        """
        alpha = 2. / (self.average + 1.)

        if data_ave is None or len(data_ave) != len(data):
            data_ave=np.array([1e-10] * len(data))

        new_data_ave = data_ave + alpha * (data - data_ave)

        return new_data_ave

    def draw_curve(self, count, data, channel, draw_trig_mark=False):
        """
        Draw a waveform curve

        :param index:  count of curve for plotting
        :param data:   curve data for plotting
        :param draw_trig_mark: flag whether drawing trigger mark
        :return:
        """

        data_len = len(data)
        if data_len == 0:
            return
        
        # in case on time reference in PV, we declare sample period to sec per sample.
        # 1 second per sample as initial
        sample_period = 1.0
        # time array
        time_array = None

        if self.current_xaxes != "None" and len(self.data.get(self.current_xaxes,[])) == data_len:
            # TODO later to support: frequency field & sample period as time reference
            # TODO need to handle multiple waveform plotting with different data length
            # Currently, support time only
            sample_period = np.diff(self.data[self.current_xaxes]).mean()
            time_array = self.data[self.current_xaxes]

        if self.diff:
            d = np.diff(data)

        xf = yf = None
        if self.fft or self.psd:
            if data_len == 0:
                return
            mode = "fft" if self.fft else "psd"
            xf, yf = self.calculate_ftt(data, sample_period, mode, self.fft_filter)

        if self.histogram and not self.psd and not self.fft:
            self.curves[count].opts['stepMode'] = True
        else:
            self.curves[count].opts['stepMode'] = False

        # Exponential moving average
        if self.average > 1:
            if yf is not None:
                yf = self.exponential_moving_average(yf, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = yf
            else:
                data = self.exponential_moving_average(data, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = data

        # Exponential moving average
        if self.average > 1:
            if yf is not None:
                yf = self.exponential_moving_average(yf, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = yf
            else:
                data = self.exponential_moving_average(data, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = data

        # Draw curve for the fft or psd mode
        if self.fft or self.psd:
            self.curves[count].setData(xf, yf)

        # Calculate histogram bins and draw the curve
        elif self.histogram and not self.psd and not self.fft:
            d = self.filter(data)
            y, x = np.histogram(d, bins=self.bins)
            self.curves[count].setData(x, y)

        # Draw waveform where X axis are indexes of data array which is drawn on Y axis
        elif time_array is None:
            data_to_plot = self.filter(data) + channel.dc_offset
            self.curves[count].setData(data_to_plot)
            self.__handle_trigger_marker__(draw_trig_mark, self.trigger.display_trigger_index(), [data_to_plot.min(), data_to_plot.max()])

        # Draw time_array on the X axis and data on Y
        else:
            d, t = self.filter(data, time_array)
            data_to_plot = d + channel.dc_offset
            self.curves[count].setData(t - t[0], data_to_plot)
            self.__handle_trigger_marker__(draw_trig_mark, self.trigger.trigger_timestamp - time_array[0], [data_to_plot.min(), data_to_plot.max()])

    def __handle_trigger_marker__(self, draw_trig_mark, marktime, mark_size):
        """
        Draw vertical line trigger mark at the specified marktime.

        :param draw_trig_mark: (bool) Flag if trigger line should be drawn.
        :param marktime: (int) Sample number at which the trigger mark happened.
        :param mark_size: (Array [int, int]) Y range for the trigger mark.
        :return:
        """
        if self.trigger.is_triggered():
            if draw_trig_mark:
                # Add trigger marker on plotting
                pvnames = [ c.pvname for c in self.channels ]
                min_value = max_value = None
                if self.single_axis:
                    for field, value in self.data.items():
                        if field in pvnames:
                            a_min = np.amin(value).item()
                            a_max = np.amax(value).item()
                            min_value = a_min if min_value is None else min(min_value, a_min)
                            max_value = a_max if max_value is None else max(max_value, a_max)
                    mark_size = [min_value, max_value]
                else:
                    mark_size = mark_size
                # Create a new pen each plot to ensure line width is 1
                # Otherwise, line can disappear when increasing size of data to plot
                self.trigMarker.setPen(pyqtgraph.mkPen(color='r', width=1))
                self.trigMarker.setData(np.array([marktime, marktime]),
                                        np.array(mark_size))
        else:
            self.trigMarker.clear()
            
    def update_drawing(self):
        """
        Update display plotting

        :return:
        """

        if self.first_data or not self.plotting_started:
            # TODO this is to help to over come race condition. To be done in a better way later
            # self.signal()
            return

        # Do not update the drawing if image is frozen
        if self.is_freeze:
            return

        # Take ownership the self.data variable which holds all the waveforms
        self.wait()

        if self.sampling_mode:
            max_samples = min(max([0] + [len(v) for v in self.data.values()]) + 1, self.max_length)
            
            for k, v in self.sample_data.items():
                data = np.append(self.data.get(k, []), v)[-self.max_length:]
                #
                # If started collecting PV data after other PVs, pad any data points before collection
                # with zeros
                #
                if max_samples > len(data):
                    data = np.pad(data, (max_samples - len(data), 0), 'constant', constant_values=0)
                self.data[k] = data;
        # Iterate over the names (channels) selected on the GUI and draw a line for each
        count = 0
        draw_trig_mark = True
        for idx, ch in enumerate(self.channels):
            if ch.pvname != "None":
                data = self.data.get(ch.pvname)
                if data is not None:
                    self.draw_curve(count, data, ch, draw_trig_mark)
                    draw_trig_mark = False
                count = count + 1

        # Axis scaling
        if self.first_run or self.new_plot:
            self.do_autoscale()
            self.reset_xrange()
        elif self.new_buffer:
            self.reset_xrange()

        self._setup_ticks()
            
        self.first_run = False
        self.new_plot = False
        self.new_buffer = False        

        self.trigger.finish_drawing()

        self.mouse_over.update_textbox()
        
        self.update_fps()

        # Release the ownership on the self.data
        self.signal()

    def update_fps(self):
        """
        Update fps statistics

        :return:
        """
        now = time.time()
        dt = now - self.lastTime
        self.lastTime = now
        if self.fps is None:
            self.fps = 1.0 / dt
        else:
            s = np.clip(dt * 3., 0, 1)
            self.fps = self.fps * (1 - s) + (1.0 / dt) * s

    def notify_plotting_started(self, val):
        self.plotting_started = val
        if self.trigger.trigger_mode:
            if val:
                self.trigger.reset()
                self.trigger.connect_to_callback()
            else:
                self.trigger.disconnect_to_callback()



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Various custom UI components.
"""
import pyqtgraph as pg
from PyQt5.QtCore import QRect, Qt
from PyQt5.QtGui import QPainter, QBrush, QPalette, QPen
from PyQt5.QtWidgets import QRubberBand, QWidget

class TransparentRubberBand(QRubberBand):
    def __init__(self, shape, parent):
        QRubberBand.__init__(self, shape, parent)

    def paintEvent(self, event):
        pal = QPalette()
        pal.setBrush(QPalette.Highlight, QBrush(Qt.black))
        self.setPalette(pal)
        pen = QPen()
        pen.setStyle(Qt.DashLine)
        pen.setColor(Qt.darkGreen)
        pen.setWidth(2)
        painter = QPainter(self)
        painter.setOpacity(1)
        painter.setPen(pen)
        rectangle = QRubberBand.geometry(self)
        rectangle2 = QRect(0,0,rectangle.width()-1, rectangle.height()-1)
        painter.drawRect(rectangle2)
        painter.end()

class RoiMidLines(QWidget):
    def __init__(self, xleft, yleft, xtop, ytop, parent=None):
        QWidget.__init__(self, parent)
        self.xleft = int(xleft)
        self.yleft = int(yleft)
        self.xtop = int(xtop)
        self.ytop = int(ytop)

    def paintEvent(self, event):
        pal = QPalette()
        pal.setBrush(QPalette.Highlight, QBrush(Qt.black))
        self.setPalette(pal)
        pen = QPen()
        pen.setStyle(Qt.DashLine)
        pen.setColor(Qt.darkGreen)
        pen.setWidth(2)
        painter = QPainter(self)
        painter.setPen(pen)
        painter.drawLine(0, self.yleft, self.xleft, self.yleft)
        painter.drawLine(self.xtop, 0, self.xtop, self.ytop)
        painter.end()



# -*- coding: utf-8 -*-

from .scope_config_base import ScopeConfigureBase
import logging
import re

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""


class Configure(ScopeConfigureBase):
    """
    Scope application configuration panel settings
    """
    def __init__(self, params, **kwargs):
        """

        :param params: parameters parsed from command line and configuration file
        :param pvs: pv name dictionary, format: {"alias": "PV:Name"}
        """
        super().__init__(params, **kwargs)
        self.pvs = kwargs.get("pv", None)

        self.counts = 4
        self.default_color = ['#FFFF00', '#FF00FF', '#55FF55', '#00FFFF', '#5555FF',
                              '#5500FF', '#FF5555', '#0000FF', '#FFAA00', '#000000']
        
        self.default_arrayid = kwargs.get("arrayid", "None")
        self.default_xaxes = kwargs.get("xaxes", "None")

        
    def assemble_channel(self, section=None):
        """
        Assemble channel information for plotting

        :param section:
        :return:
        """
        # get channel counts to display, 4 by default
        if section is None:
            self.counts = 4
        else:
            self.counts = section.getint('COUNT', 4)

        channel = []

        #Read channel information.  Channel order is
        #determined by order in config
        chan_cfg_lookup = {}
        
        if section:
            for k, v in section.items():
                if bool(re.match('chan', k, re.I)):
                    ch, param = k.lower().split('.')
                    chan_cfg_lookup[ch] = chan_cfg_lookup.get(ch, {})
                    chan_cfg_lookup[ch][param] = v

        chan_cfgs = list(chan_cfg_lookup.values())
        if len(chan_cfgs) > self.counts:
            self.counts = len(chan_cfgs)

        if self.counts > 10:
            # limit max channel to display
            self.counts = 10

        
        for i in range(self.counts):
            default_cfg = {
                'field' : 'None',
                'dcoffset': 0.0
            }
            
            chcfg = chan_cfgs[i] if len(chan_cfgs) > i else default_cfg
            field = chcfg.get('field', default_cfg['field'])
            dcoffset = float(chcfg.get('dcoffset', default_cfg['dcoffset']))

            channel.append(
                {"name": "Channel %s" % (i + 1),
                 "type": "group",
                 "children": [
                     {
                         "name": "Color",
                         "type": "color",
                         "value": self.default_color[i],
                         "readonly": True
                     },
                     {"name": "Field", "type": "list", "limits": [field], "value": field},
                     {"name": "DC offset", "type": "float", "value": dcoffset},
                     {"name": "Axis location", "type": "list", "limits": {
                         "Left" : "left",
                         "Right" : "right",
                     }, "value" : "Left"},
                 ]
                 }
            )

        return channel

    def assemble_display(self, section=None):
        display = super().assemble_display(section=section, app_section_key="SCOPE", default_autoscale=False)
        #FIXME: move this to base class once get mouseover support in striptool
        children = display['children']
        newchildren = [
            { "name" : "Mouse Over", "type": "bool", "value": False }
            ]
        children.extend(newchildren)
        
        return display

    def assemble_acquisition(self, section=None):
        buffer_unit = 'samples'
        try:
            buffer_unit = section["BUFFERUNIT"]
        except:
            pass

        if buffer_unit.lower() not in ["samples", "objects"]:
            buffer_unit = 'samples'

        buffer_unit = buffer_unit.title()

        acquisition = super().assemble_acquisition(section, buffer_unit=buffer_unit)
        children = acquisition['children']

        start = False
        if section:
            start = section.getboolean('ConnectOnStart', False)

        pv = None
        if self.pvs is not None:
            pv = list(self.pvs.values())[0]

        if section:
            if pv is None:
                # it means PV map is not specified from command line
                # get one from configuration
                pv = pv if pv is not None else section.get("PV", None)
                if pv is not None:
                    # if PV is available by default
                    self.pvs = {pv: pv}

        newchildren = [
            {"name": "Buffer Unit", "type": "list", "limits": ["Samples", "Objects"],
             "value": buffer_unit},
            {"name": "PV", "type": "str", "value": pv},
            {"name": "PV status", "type": "str", "value": "Disconnected", "readonly": True},
            {"name": "Start", "type": "bool", "value": start}
        ]

        children.extend(newchildren)
        
        return acquisition

    def assemble_statistics(self):
        stats = super().assemble_statistics()
        children = stats['children']
        children.append( {"name": "Avg Samples/Obj", "type": "float", "value": 0, "readonly":True, "decimals":20})
        return stats
    
    def assemble_config(self, section=None):
        # Assemble extra configuration information for plotting
        # which is ArrayId selection, and x axes
        id_value = ["None"]
        if self.default_arrayid != "None":
            id_value.append(self.default_arrayid)
        axes = ["None"]
        self.default_xaxes = section.get("XAXES", "None") if section else "None"
        if self.default_xaxes != "None":
            axes.append(self.default_xaxes)

        self.default_major_tick = section.getint('MAJORTICKS', 0) if section else 0
        self.default_minor_tick = section.getint('MINORTICKS', 0) if section else 0
        
        cfg = {"name": "Config",
               "type": "group",
               "expanded": True,
               "children": [
                   {"name": "ArrayId", "type": "list", "limits": id_value, "value": self.default_arrayid},
                   {"name": "X Axes", "type": "list", "limits": axes, "value": self.default_xaxes},
                   {"name": "Major Ticks", "type": "int", "value": self.default_major_tick, 'decimals':20},
                   {"name": "Minor Ticks", "type": "int", "value": self.default_minor_tick, 'decimals':20},
                   {"name": "Extra Display Fields", "type": "checklist", "limits":[], "expanded": False},
                   {"name": "MO Disp Location", "type": "list", "limits":['top-right', 'bottom-right', 'bottom-left'], "value" : "bottom-right"}
                   
                   ]
               }
        return cfg

    def parse(self):
        """

        :return:
        """
        try:
            sections = self.params["SCOPE"]["SECTION"].split(",")
            for idx, section in enumerate(sections):
                # remove unnecessary space
                sections[idx] = section.strip() 
            if "ACQUISITION" in sections:
                acquisition = self.assemble_acquisition(self.params["ACQUISITION"])
            else:
                acquisition = self.assemble_acquisition()
            if "DISPLAY" in sections:
                display = self.assemble_display(self.params["DISPLAY"])
            else:
                display = self.assemble_display()
            if "CHANNELS" in sections:
                channel = self.assemble_channel(self.params["CHANNELS"])
            else:
                channel = self.assemble_channel()
            if "TRIGGER" in sections:
                trigger = self.assemble_trigger(self.params["TRIGGER"])
            else:
                trigger = self.assemble_trigger()
            if "CONFIG" in sections:
                cfg = self.assemble_config(self.params["CONFIG"])
            else:
                cfg = self.assemble_config()
                                           
        except KeyError as e:
            if str(e) not in ["'SCOPE'"]:
                logging.getLogger().warning('Key %s not found in config' %  (str(e)))
            acquisition = self.assemble_acquisition()
            display = self.assemble_display()
            channel = self.assemble_channel()
            trigger = self.assemble_trigger()
            cfg = self.assemble_config()

        # line up in order
        paramcfg = [acquisition, trigger, display, cfg]
        for ch in channel:
            paramcfg.append(ch)
        statistics = self.assemble_statistics()
        
        paramcfg.append(statistics)

        return paramcfg



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2020 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Unit tests for Imagev

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import os
import sys
import unittest
import unittest.mock as mock
from pvaccess import PvObject
from pvaccess import NtNdArray
import pvaccess as pva
import numpy as np
from pyqtgraph.Qt import QtWidgets
from pyqtgraph import QtCore

from c2dataviewer.imagev import ImageSettingsDialog, WarningDialog
from c2dataviewer.imagev import ImageController
from c2dataviewer.imagev import ImageWindow

from .helper import create_image
from .helper import setup_qt_app

os.environ["QT_QPA_PLATFORM"] = "offscreen"

class TestImageDisplay(unittest.TestCase):

    def setUp(self):
        """
        Build the environment for each unit test case.
        This method is called before each test.

        :return:
        """

        # Create Qt application
        setup_qt_app()
        
        # Create ImageWindow and het the imageWidget instance
        self.window = ImageWindow()
        self.imageWidget = self.window.imageWidget

        # Build GUI elements
        w = ImageWindow(None)
        datasource = mock.Mock()
        w.imageWidget.set_datasource(datasource)
        image_settings_dialog = ImageSettingsDialog(None)
        warning = WarningDialog(None)

        self.ic = ImageController(w, IMAGE_SETTINGS_DIALOG=image_settings_dialog, WARNING=warning,
                PV={'test', 'test'}, timer=QtCore.QTimer(), data=None)

    def tearDown(self):
        """
        Tear down the environment after each test case.
        This mentod is called after each test.

        :return:
        """

    def test_blackSetting(self):
        """
        Test if setting black value behaves correctly

        :return:
        """

        # Test default values
        self.assertEqual(0, self.ic._win.imageBlackSlider.value())
        self.assertEqual(0, self.ic._win.imageBlackSlider.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageBlackSlider.maximum())
        self.assertEqual(0, self.ic._win.imageBlackSpinBox.value())
        self.assertEqual(0, self.ic._win.imageBlackSpinBox.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageBlackSpinBox.maximum())


        # Set black to 100
        self.ic.updateGuiBlack(100)
        self.assertEqual(100, self.ic._win.imageBlackSlider.value())
        self.assertEqual(0, self.ic._win.imageBlackSlider.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageBlackSlider.maximum())
        self.assertEqual(100, self.ic._win.imageBlackSpinBox.value())
        self.assertEqual(0, self.ic._win.imageBlackSpinBox.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageBlackSpinBox.maximum())

        # Set limits
        self.ic.changeimageBlackLimits(-10, 50)
        self.assertEqual(50, self.ic._win.imageBlackSlider.value())
        self.assertEqual(-10, self.ic._win.imageBlackSlider.minimum())
        self.assertEqual(50, self.ic._win.imageBlackSlider.maximum())
        self.assertEqual(50, self.ic._win.imageBlackSpinBox.value())
        self.assertEqual(-10, self.ic._win.imageBlackSpinBox.minimum())
        self.assertEqual(50, self.ic._win.imageBlackSpinBox.maximum())

        # Set limits bigger than slider allows (test factor implementation)
        self.ic.changeimageBlackLimits(0, self.ic.SLIDER_MAX_VAL * 2)
        self.ic.updateGuiBlack(self.ic.SLIDER_MAX_VAL * 1.5)
        self.assertAlmostEqual(self.ic.SLIDER_MAX_VAL * 1.5, self.ic._win.imageBlackSlider.value() / self.ic._imageBlackSliderFactor, delta=1)
        self.assertEqual(0, self.ic._win.imageBlackSlider.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageBlackSlider.maximum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL * 1.5, self.ic._win.imageBlackSpinBox.value())
        self.assertEqual(0, self.ic._win.imageBlackSpinBox.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL * 2, self.ic._win.imageBlackSpinBox.maximum())

    def test_whiteSetting(self):
        """
        Test if setting white value behaves correctly

        :return:
        """

        # Test default values
        self.assertEqual(1, self.ic._win.imageWhiteSlider.value())
        self.assertEqual(1, self.ic._win.imageWhiteSlider.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageWhiteSlider.maximum())
        self.assertEqual(1, self.ic._win.imageWhiteSpinBox.value())
        self.assertEqual(1, self.ic._win.imageWhiteSpinBox.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageWhiteSpinBox.maximum())


        # Set White to 100
        self.ic.updateGuiWhite(100)
        self.assertEqual(100, self.ic._win.imageWhiteSlider.value())
        self.assertEqual(1, self.ic._win.imageWhiteSlider.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageWhiteSlider.maximum())
        self.assertEqual(100, self.ic._win.imageWhiteSpinBox.value())
        self.assertEqual(1, self.ic._win.imageWhiteSpinBox.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageWhiteSpinBox.maximum())

        # Set limits
        self.ic.changeimageWhiteLimits(-10, 50)
        self.assertEqual(50, self.ic._win.imageWhiteSlider.value())
        self.assertEqual(-10, self.ic._win.imageWhiteSlider.minimum())
        self.assertEqual(50, self.ic._win.imageWhiteSlider.maximum())
        self.assertEqual(50, self.ic._win.imageWhiteSpinBox.value())
        self.assertEqual(-10, self.ic._win.imageWhiteSpinBox.minimum())
        self.assertEqual(50, self.ic._win.imageWhiteSpinBox.maximum())

        # Set limits bigger than slider allows (test factor implementation)
        self.ic.changeimageWhiteLimits(0, self.ic.SLIDER_MAX_VAL * 2)
        self.ic.updateGuiWhite(self.ic.SLIDER_MAX_VAL * 1.5)
        self.assertAlmostEqual(self.ic.SLIDER_MAX_VAL * 1.5, self.ic._win.imageWhiteSlider.value() / self.ic._imageWhiteSliderFactor, delta=1)
        self.assertEqual(0, self.ic._win.imageWhiteSlider.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL, self.ic._win.imageWhiteSlider.maximum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL * 1.5, self.ic._win.imageWhiteSpinBox.value())
        self.assertEqual(0, self.ic._win.imageWhiteSpinBox.minimum())
        self.assertEqual(self.ic.SLIDER_MAX_VAL * 2, self.ic._win.imageWhiteSpinBox.maximum())

    def test_zoomInformation(self):
        """
        Test if zoom information is properly displayed on the GUI

        :return:
        """

        # Setup image inside the ImagePlotWidget
                # Build test image
        arrayValue = [
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            ]

        data = create_image(1, arrayValue, nx=10, ny=10, color_mode=0)


        # Display original image
        self.ic._win.imageWidget._ImagePlotWidget__update_dimension(data)
        self.ic._win.imageWidget.display(data)
        self.ic._win.imageWidget.data = data

        # Update the GUI
        self.ic.updateStatus()

        # Check the information on the gui
        self.assertEqual("10", self.ic._win.lblXsize.text().strip())
        self.assertEqual("10", self.ic._win.lblYsize.text().strip())
        self.assertEqual("0", self.ic._win.deadPixel.text().strip())
        self.assertEqual("255.0", self.ic._win.maxPixel.text().strip())
        self.assertEqual("0.0", self.ic._win.minPixel.text().strip())

        # Set zoom region
        self.ic._win.imageWidget.set_zoom_region(4, 3, 5, 8)

        # Update the GUI
        self.ic.updateStatus()

        # Check the information on the gui
        self.assertEqual("10 (5|4-9)", self.ic._win.lblXsize.text().strip())
        self.assertEqual("10 (7|3-10)", self.ic._win.lblYsize.text().strip())
        self.assertEqual("0 (0)", self.ic._win.deadPixel.text().strip())
        self.assertEqual("255.0 (76.0)", self.ic._win.maxPixel.text().strip())
        self.assertEqual("0.0 (23.0)", self.ic._win.minPixel.text().strip())

        # Reset zoom
        self.ic._win.imageWidget.reset_zoom()

        # Update the GUI
        self.ic.updateStatus()

        # Check the information on the gui
        self.assertEqual("10", self.ic._win.lblXsize.text().strip())
        self.assertEqual("10", self.ic._win.lblYsize.text().strip())
        self.assertEqual("0", self.ic._win.deadPixel.text().strip())
        self.assertEqual("255.0", self.ic._win.maxPixel.text().strip())
        self.assertEqual("0.0", self.ic._win.minPixel.text().strip())



############################################
# Test statistics
############################################

    def test_statistics(self):
        """
        Test statistics calculations.

        :return:
        """
        # Constants
        NX = 10
        NY = 10
        COLOR_MODE = 0
        N_IMAGES = 20 # For each rate

        # Build image
        self.imageWidget.x = NX
        self.imageWidget.y = NY

        # Generate images
        array_value = [
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            255, 0, 77, 54, 23, 76, 34, 65, 34, 65,
            ]
        np_value = np.array(array_value, dtype=np.uint8)
        image_counter = 0

        # Calculate init statistics
        self.ic.calculate_statistics()
        self.assertAlmostEqual(0.0, self.ic._win.imageWidget.MB_received, delta=0.0001)
        self.assertEqual(0, self.ic._win.imageWidget.frames_displayed)
        self.assertAlmostEqual(0, self.ic.fps_current, delta=0.1)
        self.assertAlmostEqual(0, self.ic.fps_average, delta=0.1)
        self.assertAlmostEqual(0, self.ic.fps_current_received, delta=0.1)
        self.assertAlmostEqual(0, self.ic.fps_average_received, delta=0.1)
        self.assertEqual(0, self.ic.frames_missed_current)
        self.assertAlmostEqual(0.0, self.ic.frames_missed_average, delta=0.5)
        self.assertEqual(0, self.ic._win.imageWidget.frames_missed)
        self.assertEqual(0, self.ic._win.imageWidget.minVal)
        self.assertEqual(0, self.ic._win.imageWidget.maxVal)

        # Generate images
        for i in range(N_IMAGES):
            image_counter += 1
            if i % 3 == 0:
                continue
            image = create_image(image_counter, np_value, nx=NX, ny=NY, color_mode=COLOR_MODE,  time_stamp=image_counter)
            self.ic._win.imageWidget._ImagePlotWidget__update_dimension(image)
            self.ic._win.imageWidget.display(image)

        # Calculate statistics
        self.ic.calculate_statistics()
        self.assertAlmostEqual(0.0012, self.ic._win.imageWidget.MB_received, delta=0.0001)
        self.assertEqual(12, self.ic._win.imageWidget.frames_displayed)
        self.assertEqual(7, self.ic._win.imageWidget.frames_missed)
        self.assertEqual(min(array_value), self.ic._win.imageWidget.minVal)
        self.assertEqual(max(array_value), self.ic._win.imageWidget.maxVal)

        # Generate images
        for i in range(N_IMAGES):
            image_counter += 1
            if i % 3 == 0:
                continue
            image = create_image(image_counter, np_value, nx=10, ny=10, color_mode=0,  time_stamp=image_counter)
            self.ic._win.imageWidget._ImagePlotWidget__update_dimension(image)
            self.ic._win.imageWidget.display(image)

        # Calculate statistics
        self.ic.calculate_statistics()
        self.assertAlmostEqual(0.0025, self.ic._win.imageWidget.MB_received, delta=0.0001)
        self.assertEqual(25, self.ic._win.imageWidget.frames_displayed)
        self.assertEqual(14, self.ic._win.imageWidget.frames_missed)
        self.assertEqual(min(array_value), self.ic._win.imageWidget.minVal)
        self.assertEqual(max(array_value), self.ic._win.imageWidget.maxVal)

        # Generate images
        for i in range(N_IMAGES):
            image_counter += 1
            if i % 3 == 0:
                continue
            image = create_image(image_counter, np_value, nx=10, ny=10, color_mode=0,  time_stamp=image_counter)
            self.ic._win.imageWidget._ImagePlotWidget__update_dimension(image)
            self.ic._win.imageWidget.display(image)

        # Calculate statistics
        self.ic.calculate_statistics()
        self.assertAlmostEqual(0.0038, self.ic._win.imageWidget.MB_received, delta=0.0001)
        self.assertEqual(38, self.ic._win.imageWidget.frames_displayed)
        self.assertEqual(21, self.ic._win.imageWidget.frames_missed)
        self.assertEqual(min(array_value), self.ic._win.imageWidget.minVal)
        self.assertEqual(max(array_value), self.ic._win.imageWidget.maxVal)



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""
import datetime
from pyqtgraph import Qt
from pyqtgraph import QtCore
import time
from pyqtgraph.functions import mkPen
from pyqtgraph import PlotWidget
from PyQt5 import QtWidgets

from ..view.image_definitions import COLOR_MODE_MONO, COLOR_MODES
from ..model import ConnectionState

class ImageController:

    SLIDER_MAX_VAL = 1_000_000_000
    SLIDER_MIN_VAL = -SLIDER_MAX_VAL
    SPINNER_MAX_VAL = 1.7976931348623157e+308
    SPINNER_MIN_VAL = -SPINNER_MAX_VAL

    HIDE_CONTROL_TEXT = 'Hide Control Panel'
    SHOW_CONTROL_TEXT = 'Show Control Panel'
    HIDE_XY_INTENSITY_TEXT = 'Hide Coordinates and Intensity'
    SHOW_XY_INTENSITY_TEXT = 'Show Coordinates and Intensity'
    RESET_ZOOM_TEXT = 'Reset Zoom'
    ENABLE_ROI_MODE_TEXT = 'Enable ROI Mode'
    DISABLE_ROI_MODE_TEXT = 'Disable ROI Mode'

    def __init__(self, widget, **kargs):
        """

        :param widget:
        :param kargs:
        """
        self._win = widget
        if self._win is None:
            raise RuntimeError("Widget is unknown")

        cameras = kargs.get("PV", None)

        self.profilesWidgetMutex = QtCore.QMutex()
        self.profilesWidgetDisplayed = False

        self.datareceiver = kargs.get('data', None)

        self._win.pvPrefix.setEditable(True)
        self._win.pvPrefix.setInsertPolicy(Qt.QtWidgets.QComboBox.InsertAtBottom)
        if cameras is not None:
            self._win.pvPrefix.addItems(cameras)
        self._win.pvPrefix.currentIndexChanged.connect(self.camera_changed)
        self._win.imageWidget.connection_changed_signal.connect(self.connection_changed)
        
        self._image_settings_dialog  = kargs.get("IMAGE_SETTINGS_DIALOG", None)
        self._warning = kargs.get("WARNING", None)

        # GUI styles
        self._inputTypeDefaultStyle = self._win.tbValidInput.styleSheet()

        # Image and zoom
        self._win.lblXsize.setToolTip("Number of pixels image has in X direction. \nIf ROI is selected, numbers in parentheses \nshow the size and range of displayed pixels.")
        self._win.lblYsize.setToolTip("Number of pixels image has in Y direction. \nIf ROI is selected, numbers in parentheses \nshow the size and range of displayed pixels.")
        self._win.resetZoomButton.clicked.connect(lambda: self._callback_reset_zoom_button())

        # Black control
        self._win.imageBlackSlider.valueChanged.connect(lambda: self._callback_black_changed_slider())
        self._win.imageBlackSpinBox.valueChanged.connect(lambda: self._callback_black_changed_spin())
        self.changeimageBlackLimits(0, self.SLIDER_MAX_VAL)
        self._imageBlackSliderFactor = 1
        self._win.imageWidget.set_BlackLimitsCallback(self.changeimageBlackLimits)
        self._win.imageWidget.set_BlackCallback(self.updateGuiBlack)

        # White control
        self._win.imageWhiteSlider.valueChanged.connect(lambda: self._callback_white_changed_slider())
        self._win.imageWhiteSpinBox.valueChanged.connect(lambda: self._callback_white_changed_spin())
        self.changeimageWhiteLimits(1, self.SLIDER_MAX_VAL)
        self._imageWhiteSliderFactor = 1
        self._win.imageWidget.set_WhiteLimitsCallback(self.changeimageWhiteLimits)
        self._win.imageWidget.set_WhiteCallback(self.updateGuiWhite)

        # Auto adjust black/white
        self._win.imageAutoAdjust.clicked.connect(lambda: self.auto_levels_cal())
        self._win.imageWidget.set_getBlackWhiteLimits(self.getimageBlackLimits, self.getimageWhiteLimits)

        # Moving average
        self._win.cbEnableMovingAverage.setChecked(False)
        self._win.sbMovingAverageFrames.hide()
        self._win.lblMovingAverageFrames.hide()
        self._win.lblProfilesSpacer.hide()
        self._win.sbMovingAverageFrames.setMinimum(2)
        self._win.sbMovingAverageFrames.setValue(10)
        self._win.sbMovingAverageFrames.setMaximum(10000)
        self._win.sbMovingAverageFrames.setToolTip("Number of frames to be used for calculating moving image average. ")

        # Set limits on the dialog widgets
        self._image_settings_dialog.blackMin.setMinimum(self.SPINNER_MIN_VAL)
        self._image_settings_dialog.blackMin.setMaximum(self.SPINNER_MAX_VAL)
        self._image_settings_dialog.blackMax.setMinimum(self.SPINNER_MIN_VAL)
        self._image_settings_dialog.blackMax.setMaximum(self.SPINNER_MAX_VAL)
        self._image_settings_dialog.whiteMin.setMinimum(self.SPINNER_MIN_VAL)
        self._image_settings_dialog.whiteMin.setMaximum(self.SPINNER_MAX_VAL)
        self._image_settings_dialog.whiteMax.setMinimum(self.SPINNER_MIN_VAL)
        self._image_settings_dialog.whiteMax.setMaximum(self.SPINNER_MAX_VAL)
        self._image_settings_dialog.displayQueueSize.setMinimum(self._win.imageWidget.MIN_DISPLAY_QUEUE_SIZE)
        self._image_settings_dialog.displayQueueSize.setMaximum(self._win.imageWidget.MAX_DISPLAY_QUEUE_SIZE)
        self._image_settings_dialog.sbCpuLimit.setMinimum(10)
        self._image_settings_dialog.sbCpuLimit.setMaximum(400)
        self._image_settings_dialog.sbNetworkLimit.setMinimum(10)
        self._image_settings_dialog.sbNetworkLimit.setMaximum(1250) # 1250 megabyte = 10 gigabit
        self._image_settings_dialog.sbServerQueueSize.setMinimum(0)
        self._image_settings_dialog.sbServerQueueSize.setMaximum(10000)

        # Configure settings window tooltips
        self._image_settings_dialog.sbDeadPixelThreshold.setToolTip("Dead pixel threshold. \nPixel values above this setting are counted as 'dead'.")
        self._image_settings_dialog.sbEmbeddedDataLength.setToolTip("Number of pixels at the beggining of the image that hold the embedded data. \nThese pixels are not included in the statistics calculations.")
        self._image_settings_dialog.sbCpuLimit.setToolTip("Maximum allowed usage of the CPU. \n100% = 1 full CPU core.")
        self._image_settings_dialog.sbNetworkLimit.setToolTip("Maximum allowed usage of the network in MB/s.")
        self._image_settings_dialog.sbServerQueueSize.setToolTip("Maximum number of frames to be queued on the server side.")

        # Settings dialog callbacks
        self._win.btnSettings.clicked.connect(lambda: self._callback_adjust_image_settings())
        self._image_settings_dialog.okButton.clicked.connect(lambda: self._callback_accept_new_image_settings())
        self._image_settings_dialog.cancelButton.clicked.connect(lambda: self._callback_cancel_new_image_settings())

        # Frame DAQ control
        self._framerates = {'1 Hz': 1, '2 Hz': 2, '5 Hz': 5, '10 Hz': 10, '20 Hz': 20, '30 Hz': 30, 'Full IOC Rate': -1} #Full IOC rate must be on the last place
        self._win.iocRate.addItems(self._framerates.keys())
        self._win.iocRate.setCurrentIndex(2)
        self.frameRateChanged()
        self._win.iocRate.currentIndexChanged.connect(lambda: self.frameRateChanged())
        self._win.freeze.stateChanged.connect(lambda: self._callback_freeze_changed())


        # Min/max channel select
        self.colorChannels = {
            0 : "Red",
            1 : "Green",
            2 : "Blue",
        }
        self.maxChannel = self.colorChannels[0]
        self.minChannel = self.colorChannels[0]
        self._win.maxPixelChannel.addItems(list(self.colorChannels.values()))
        self._win.minPixelChannel.addItems(list(self.colorChannels.values()))
        self._win.maxPixelChannel.setToolTip("Selection for which color channel the value is displayed.")
        self._win.minPixelChannel.setToolTip("Selection for which color channel the value is displayed.")

        # Add tooltips for statistics
        self._win.tbValidInput.setToolTip("Image data type.")
        self._win.runtime.setToolTip("Total time data viewer has been running.")
        self._win.maxPixel.setToolTip("Maximum value in the image. In color modes RGB values are shown. \nIf ROI is selected, value in the \nparentheses apply for the displayed area.")
        self._win.minPixel.setToolTip("Minimum value in the image. In color modes RGB values are shown. \nIf ROI is selected, value in the \nparentheses apply for the displayed area.")
        self._win.frameRateCurrAvg.setToolTip("Current / average display Frames Per Second.")
        self._win.receiveRateCurrAvg.setToolTip("Current / average receive Frames Per Second.")
        self._win.nFrames.setToolTip("Total number of frames processed/displayed. \nSome frames can be skipped for drawing if the painter can not keep up.")
        self._win.nMissedFramesCurrAvg.setToolTip("Current / average missed frames per second.")
        self._win.nMissedFrames.setToolTip("Total number of missed frames.")
        self._win.deadPixel.setToolTip("Number of pixels that exceed \nthe dead pixel threshold. \nIf ROI is selected, value in the \nparentheses apply for the displayed area. \nThis can be configured under 'Settings'.")
        self._win.cpuUsage.setToolTip("CPU usage of the data viewer.")
        self._win.netLoad.setToolTip("Network usage of data viewer.")
        self._win.netReceived.setToolTip("Total number of mega bytes received")

        # Reset statistics
        self._win.resetStatistics.clicked.connect(lambda: self.reset_statistics())

        self._warning.warningConfirmButton.clicked.connect(lambda: self.acceptWarning())


        # Statistics variables
        self.reset_statistics()

        # Statistics update timer. Statistics should be updated once per second.
        self.statistics_timer = kargs.get("timer", None)
        if self.statistics_timer is None:
            raise RuntimeError("No valid timer")
        self.statistics_timer.timeout.connect(self.calculate_statistics)
        self.statistics_timer.start(1000)

        # Setup profiles
        self._win.imageWidget.setup_profiles(self._win.canvasGrid)
        self._win.cbShowProfiles.stateChanged.connect(self._callback_profiles_show_changed)
        self._win.cbShowRulers.stateChanged.connect(self._callback_profiles_show_changed)
        self._win.cbEnableMovingAverage.stateChanged.connect(self._callback_enable_moving_average_changed)
        self._win.sbMovingAverageFrames.textChanged.connect(self._callback_enable_moving_average_changed)

        self.frameRateChanged()
        self.camera_changed()

        self._imageContextMenu = QtWidgets.QMenu(self._win.imageWidget)
        self._imageContextMenuAction = QtWidgets.QAction(self.HIDE_CONTROL_TEXT, self._win.imageWidget)
        self._imageContextMenu.addAction(self._imageContextMenuAction)

        self._imageXY_IntensityDialogAction = QtWidgets.QAction(self.SHOW_XY_INTENSITY_TEXT, self._win.imageWidget)
        self._imageContextMenu.addAction(self._imageXY_IntensityDialogAction)
        self._imageResetZoomAction = QtWidgets.QAction(self.RESET_ZOOM_TEXT, self._win.imageWidget)
        self._imageContextMenu.addAction(self._imageResetZoomAction)
        self._imageRoiModeAction = QtWidgets.QAction(self.ENABLE_ROI_MODE_TEXT, self._win.imageWidget)
        self._imageContextMenu.addAction(self._imageRoiModeAction)
        
        self._win.imageWidget.right_button_clicked_signal.connect(lambda point: self.on_context_menu(point))

        self._scrollAreaWidth = 0

    def on_context_menu(self, point):
        # show context menu
        action = self._imageContextMenu.exec_(self._win.imageWidget.mapToGlobal(point))
        if not action:
            return
        if action.text() == self.HIDE_CONTROL_TEXT:
            action.setText(self.SHOW_CONTROL_TEXT)
            self._win.scrollArea.setHidden(True)
            self._scrollAreaWidth = self._win.scrollArea.parent().minimumWidth()
            self._win.scrollArea.parent().setMinimumWidth(0)
            self._win.scrollArea.parent().setMinimumHeight(0)
        elif action.text() == self.SHOW_CONTROL_TEXT:
            action.setText(self.HIDE_CONTROL_TEXT)
            self._win.scrollArea.setHidden(False)
            self._win.scrollArea.parent().setMinimumWidth(self._scrollAreaWidth)
        elif action.text() == self.SHOW_XY_INTENSITY_TEXT:
            action.setText(self.HIDE_XY_INTENSITY_TEXT)
            self._win.imageWidget.mouse_dialog.enable_mouse_dialog()
            self._win.imageWidget.setMouseTracking(True)
        elif action.text() == self.HIDE_XY_INTENSITY_TEXT:
            action.setText(self.SHOW_XY_INTENSITY_TEXT)
            self._win.imageWidget.mouse_dialog.disable_mouse_dialog()
            self._win.imageWidget.setMouseTracking(False)
        elif action.text() == self.RESET_ZOOM_TEXT:
            self._win.imageWidget.reset_zoom()
        elif action.text() == self.ENABLE_ROI_MODE_TEXT:
            action.setText(self.DISABLE_ROI_MODE_TEXT)
            self._win.imageWidget.createZoomSelectionIndicator(roi_mode=True)
        elif action.text() == self.DISABLE_ROI_MODE_TEXT:
            action.setText(self.ENABLE_ROI_MODE_TEXT)
            self._win.imageWidget.createZoomSelectionIndicator(roi_mode=False)
        self._win.imageWidget.adjustSize()

    def _callback_black_changed_slider(self):
        """
        This callback is called when user change the value on the "black" slider.

        :return:
        """
        try:
            black = self._win.imageBlackSlider.value() / self._imageBlackSliderFactor
            self._win.imageWidget.set_black(black)
            self._win.imageBlackSpinBox.blockSignals(True)
            self._win.imageBlackSpinBox.setValue(black)
            self._win.imageBlackSpinBox.blockSignals(False)
        except:
            pass
        self._win.imageBlackSlider.blockSignals(False)

    def _callback_black_changed_spin(self):
        """
        This callback is called when user change the value on the "black" spinner.

        :return:
        """
        try:
            black = self._win.imageBlackSpinBox.value()
            self._win.imageWidget.set_black(black)
            self._win.imageBlackSlider.blockSignals(True)
            self._win.imageBlackSlider.setValue(black * self._imageBlackSliderFactor)
            self._win.imageBlackSlider.blockSignals(False)
        except:
            pass

    def _callback_white_changed_slider(self):
        """
        This callback is called when user change the value on the "white" slider.

        :return:
        """
        try:
            white = self._win.imageWhiteSlider.value() / self._imageWhiteSliderFactor
            self._win.imageWidget.set_white(white)
            self._win.imageWhiteSpinBox.blockSignals(True)
            self._win.imageWhiteSpinBox.setValue(white)
            self._win.imageWhiteSpinBox.blockSignals(False)
        except:
            pass

    def _callback_white_changed_spin(self):
        """
        This callback is called when user change the value on the "white" spinner.

        :return:
        """
        try:
            white = self._win.imageWhiteSpinBox.value()
            self._win.imageWidget.set_white(white)
            self._win.imageWhiteSlider.blockSignals(True)
            self._win.imageWhiteSlider.setValue(white * self._imageWhiteSliderFactor)
            self._win.imageWhiteSlider.blockSignals(False)

        except:
            pass

    def _callback_reset_zoom_button(self):
        """
        This callback is called when user click on the "Reset Zoom" button.

        :return:
        """
        self._win.imageWidget.reset_zoom()

    def _callback_profiles_show_changed(self):
        """
        Callback used when the user on the GUI ticks or un-ticks the "Show
        the image profiles" or the "Show image rulers" buttons.

        :return:
        """
        self.profilesWidgetMutex.lock()
        try:
            profilesWidgetDisplayed = self._win.cbShowProfiles.isChecked() or self._win.cbShowRulers.isChecked()
            self._win.imageWidget.image_profile_widget.showProfiles(self._win.cbShowProfiles.isChecked())
            self._win.imageWidget.image_profile_widget.showRulers(self._win.cbShowRulers.isChecked())
            self._win.imageWidget.rulersDisplayed(self._win.cbShowRulers.isChecked())
            if profilesWidgetDisplayed == self.profilesWidgetDisplayed:
                # If widget is shown already, we are done here
                return
            self.profilesWidgetDisplayed = profilesWidgetDisplayed
            if profilesWidgetDisplayed:
                self._win.iocRate.blockSignals(True)
                self._win.iocRate.removeItem(len(self._framerates)-1)
                self._win.iocRate.blockSignals(False)
                new_freq_key = list(self._framerates.keys())[list(self._framerates.values()).index(1)] # Get key for 1 Hz
                self._win.iocRate.setCurrentText(new_freq_key)
                self._win.imageWidget.image_profile_widget.show(True)
            else:
                self._win.imageWidget.image_profile_widget.show(False)
                full_rate_key = list(self._framerates.keys())[-1]
                self._win.iocRate.addItem(full_rate_key, self._framerates[full_rate_key])

            if self._win.freeze.isChecked() and cb.isChecked():
                self._win.imageWidget.calculate_profiles(self._win.imageWidget.last_displayed_image)
                self._win.imageWidget._set_image_signal.emit()
        finally:
            self.profilesWidgetMutex.unlock()

    def _callback_enable_moving_average_changed(self):
        """
        Callback used when the user on the GUI ticks or un-ticks the
        "Enable Moving Average" button.

        :return:
        """
        moving_average_enabled = self._win.cbEnableMovingAverage.isChecked()
        n_moving_average_frames = self._win.sbMovingAverageFrames.value()
        self._win.imageWidget.enable_moving_average(moving_average_enabled, n_moving_average_frames)
        if moving_average_enabled:
            self._win.sbMovingAverageFrames.show()
            self._win.lblMovingAverageFrames.show()
            self._win.lblProfilesSpacer.show()
        else:
            self._win.sbMovingAverageFrames.hide()
            self._win.lblMovingAverageFrames.hide()
            self._win.lblProfilesSpacer.hide()

    def auto_levels_cal(self):
        """

        :return:
        """
        self._win.imageWidget.enable_auto_white()

    def frameRateChanged(self):
        """

        :return:
        """
        n = self._win.iocRate.currentIndex()
        fr = list(self._framerates.keys())[n]
        try:
            self._win.imageWidget.set_framerate(self._framerates[fr])
        except NameError:
            pass

    def _callback_freeze_changed(self):
        """
        Called when used freeze or unfreeze the image.

        :return:
        """
        self._win.imageWidget.set_freeze(self._win.freeze.isChecked())


    def _callback_adjust_image_settings(self):
        """
        This callback is called when user press the "Settings" button.
        It will open the pop screen which allow to change different types of settings.

        :return:
        """
        # Set current values
        self._image_settings_dialog.blackMin.setValue(self._win.imageBlackSpinBox.minimum())
        self._image_settings_dialog.blackMax.setValue(self._win.imageBlackSpinBox.maximum())
        self._image_settings_dialog.whiteMin.setValue(self._win.imageWhiteSpinBox.minimum())
        self._image_settings_dialog.whiteMax.setValue(self._win.imageWhiteSpinBox.maximum())
        self._image_settings_dialog.whiteMax.setValue(self._win.imageWhiteSpinBox.maximum())

        self._image_settings_dialog.displayQueueSize.setValue(self._win.imageWidget.get_display_max_queue_size())

        self._image_settings_dialog.cbEnableDeadPixels.setChecked(self._win.imageWidget.get_preferences()['DPXEnabled'])
        self._image_settings_dialog.sbDeadPixelThreshold.setValue(self._win.imageWidget.get_preferences()['DPXLimit'])
        self._image_settings_dialog.sbEmbeddedDataLength.setValue(self._win.imageWidget.get_preferences()['EmbeddedDataLen'])

        self._image_settings_dialog.cbEnableCpuLimit.setChecked(self._win.imageWidget.get_preferences()['EnableCPULimit'])
        self._image_settings_dialog.sbCpuLimit.setValue(self._win.imageWidget.get_preferences()['CPULimit'])
        self._image_settings_dialog.cbEnableNetworkLimit.setChecked(self._win.imageWidget.get_preferences()['EnableNetLimit'])
        self._image_settings_dialog.sbNetworkLimit.setValue(self._win.imageWidget.get_preferences()['NetLimit'])
        self._image_settings_dialog.cbEnableServerQueue.setChecked(self._win.imageWidget.get_preferences()['EnableServerQueue'])
        self._image_settings_dialog.sbServerQueueSize.setValue(self._win.imageWidget.get_preferences()['ServerQueueSize'])

        # Launch the dialog
        self._image_settings_dialog.exec_()


    def _callback_accept_new_image_settings(self):
        """
        This callback is called when the user confirms the new settings.

        :return:
        """
        self.changeimageBlackLimits(self._image_settings_dialog.blackMin.value(),
                                    self._image_settings_dialog.blackMax.value())
        self.changeimageWhiteLimits(self._image_settings_dialog.whiteMin.value(),
                                   self._image_settings_dialog.whiteMax.value())

        preferences = {
            'DPXEnabled' : self._image_settings_dialog.cbEnableDeadPixels.isChecked(),
            'DPXLimit' : self._image_settings_dialog.sbDeadPixelThreshold.value(),
            'EmbeddedDataLen' : self._image_settings_dialog.sbEmbeddedDataLength.value(),
            'EnableCPULimit' : self._image_settings_dialog.cbEnableCpuLimit.isChecked(),
            'CPULimit' : self._image_settings_dialog.sbCpuLimit.value(),
            'EnableNetLimit' : self._image_settings_dialog.cbEnableNetworkLimit.isChecked(),
            'NetLimit' : self._image_settings_dialog.sbNetworkLimit.value(),
            'EnableServerQueue' : self._image_settings_dialog.cbEnableServerQueue.isChecked(),
            'ServerQueueSize' : self._image_settings_dialog.sbServerQueueSize.value(),
        }
        self._win.imageWidget.set_preferences(preferences)

        self._win.imageWidget.set_display_queue_size(self._image_settings_dialog.displayQueueSize.value())

        # Force start/stop if we modified preferences that affect ioc rates
        self.frameRateChanged()

        self._image_settings_dialog.close()

    def _callback_cancel_new_image_settings(self):
        """
        This callback is called when user cancels the new settings.

        :return:
        """
        self._image_settings_dialog.reject()

    def acceptWarning(self):
        """

        :return:
        """
        self._warning.close()

    def notify_warning(self, msg):
        if self._warning is not None:
            self.acceptWarning()
            self._warning.warningTextBrowse.setText(msg)
            self._warning.show()
        
    def connection_changed(self, status, msg):
        if status == str(ConnectionState.FAILED_TO_CONNECT):
            self.notify_warning('Connection lost: ' + msg)
        
        self._win.lblConnectionStatus.setText(str(status))
        
    def camera_changed(self):
        """

        :param value:
        :return:
        """

        self.reset_statistics()
        pv = self._win.pvPrefix.currentText()
        try:
            self._win.imageWidget.camera_changed(pv)
        except (ValueError, RuntimeError):
            self.notify_warning("No data from: {}. Stop image display. \n"
                                "Please select a different channel.".
                                format(pv))

    def statistics_update(self, valuefield, value, roi_value=None, **kargs):
        """
        Update widget with a new value on the GUI. If limits are specified and value is outside them, text on the widget
        will become red.

        :param valuefield: (QWidget) Widget where status should be written. It must support "setText" and  "setStyleSheet" methods.
        :param value: (Object or Tuple/List of objects) Value which should be written to the widget. Multiple values can be passed as
                        tuple or array. If this is the case, **kargs must contain a formatting string that supports supplied number of values.
        :param kargs: (dict) Following keyword parameters are supported:
                                * "fmt" - (str) Formating string. Default value is "%.1f".
                                * "hilimit"  - (number) High limit. If the max value exceeds this value, text on the widget will become bold and red. Default is None.
                                * "lowlimit" - (number) Low limit. If the min value is less than this value, text on the widget will become bold and red. Default is None.
                                * "callback" - (bool) If this is the True and high or low limits are reached, *throttleBack* method is called. The default value is False.
        :return:
        """
        fmt = kargs.get('fmt', '%.1f')
        hilimit = kargs.get('hilimit', None)
        lolimit = kargs.get('lowlimit', None)
        callback = kargs.get('callback', False)

        if type(value) is list:
            value = value[0]
        if roi_value is not None:
            fmt = f"{fmt} ({fmt})"
            if type(value) is tuple:
                value = value + roi_value
            else:
                value = (value, roi_value)

        valuefield.setText(str(fmt) % value)
        maxValue = max(value) if isinstance(value, (list, tuple)) else value
        minValue = min(value) if isinstance(value, (list, tuple)) else value
        if hilimit is not None and maxValue > hilimit:
            valuefield.setStyleSheet('color: red; font-weight: bold')
            if callback:
                self.throttleBack()
        elif lolimit is not None and minValue < lolimit:
            valuefield.setStyleSheet('color: red; font-weight: bold')
            if callback:
                self.throttleBack()
        else:
            valuefield.setStyleSheet('color: black; font-weight: normal')

    def throttleBack(self):
        """

        :return:
        """
        selected = self._win.iocRate.currentIndex()
        newChoice = selected - 1
        if newChoice >= 0:
            self._win.iocRate.setCurrentIndex(newChoice)
            self.frameRateChanged()
            # self._win.setStatus('Requested frame rate reduced -- %s limit exceeded.' % sender.name)


    def calculate_statistics(self):
        """
        Calculate statistics.

        :return:
        """
        # Handle mouse dialog when freeze enabled
        if self._win.imageWidget._freeze and self._win.imageWidget.mouse_dialog.mouse_dialog_enabled:
            self._win.imageWidget.setup_mouse_textbox()

        # Get current time
        now = time.time()

        # Set required variabels on first call
        if self._last_time_stat_calculated is None:
            self._last_time_stat_calculated = now
            return

        # Calculate runtime
        self.runtime = now - self._start_time

        # Calculate time difference betwean calls
        delta_time = now - self._last_time_stat_calculated
        self._delta_time.append(delta_time)
        self._delta_time = self._delta_time[-3:]

        # Calculate number of frames displayed betwean calls
        delta_frames = self._win.imageWidget.frames_displayed - self._last_frames
        self._last_frames = self._win.imageWidget.frames_displayed
        self._delta_frames.append(delta_frames)
        self._delta_frames = self._delta_frames[-3:]
        try:
            self.fps_current = sum(self._delta_frames) / sum(self._delta_time)
        except: # pylint: disable=bare-except
            self.fps_current = 0
        self.fps_current = self.fps_current if self.fps_current >=0 else 0
        self.fps_average = self._win.imageWidget.frames_displayed / self.runtime

        # Calculate number of frames received betwean calls
        delta_frames_received = self._win.imageWidget.frames_received - self._last_frames_received
        self._last_frames_received = self._win.imageWidget.frames_received
        self._delta_frames_received.append(delta_frames_received)
        self._delta_frames_received = self._delta_frames_received[-3:]
        try:
            self.fps_current_received = sum(self._delta_frames_received) / sum(self._delta_time)
        except: # pylint: disable=bare-except
            self.fps_current_received = 0
        self.fps_current_received = self.fps_current_received if self.fps_current_received >=0 else 0
        self.fps_average_received = self._win.imageWidget.frames_received / self.runtime

        # Calculate number of missed frames betwean calls
        delta_missed_frames = self._win.imageWidget.frames_missed - self._last_missed_frames
        self._last_missed_frames = self._win.imageWidget.frames_missed
        self._delta_missed_frames.append(delta_missed_frames)
        self._delta_missed_frames = self._delta_missed_frames[-3:]
        try:
            self.frames_missed_current = sum(self._delta_missed_frames) / sum(self._delta_time)
        except:
            self.frames_missed_current = 0
        self.frames_missed_current = self.frames_missed_current if self.frames_missed_current >=0 else 0
        self.frames_missed_average = self._win.imageWidget.frames_missed / self.runtime

        # Calculate number of bytes received betwean calls
        delta_bytes = self._win.imageWidget.MB_received - self._last_MB_received
        self._last_MB_received = self._win.imageWidget.MB_received
        self._delta_bytes.append(delta_bytes)
        self._delta_bytes = self._delta_bytes[-3:]

        # Calculate machine resources usage
        with self._win._proc.oneshot():
            self.cpu_usage = self._win._proc.cpu_percent(None)
        try:
            self.network_usage = sum(self._delta_bytes) / sum(self._delta_time)
        except:
            self.network_usage = 0
        if self.network_usage < 0:
            self.network_usage = 0

        # Update time when statistics were updated
        self._last_time_stat_calculated = now

        # Update the GUI
        self.updateStatus()

    def reset_statistics(self):
        """
        Reset all the statistics.

        :return:
        """
        self._last_time_stat_calculated = None
        self._win.imageWidget.__last_array_id = None
        self._start_time = time.time()
        self.runtime = 0
        self._win.imageWidget.MB_received = 0.0
        self._win.imageWidget.frames_received = 0
        self._win.imageWidget.frames_displayed = 0
        self._last_frames = 0
        self._last_frames_received = 0
        self.fps_current = 0
        self.fps_current_received = 0
        self.fps_average = 0
        self.fps_average_received = 0
        self._win.imageWidget.frames_missed = 0
        self._last_missed_frames = 0
        self.frames_missed_current = 0
        self.frames_missed_average = 0
        self._last_MB_received = 0
        self.cpu_usage = 0
        self.network_usage = 0
        self._delta_frames = [0]
        self._delta_frames_received = [0]
        self._delta_missed_frames = [0]
        self._delta_bytes = [0]
        self._delta_time = [1]


    def updateStatus(self):
        """
        Update all the statistics values on the GUI.

        :return:
        """
        # Update input type
        self._win.tbValidInput.setText(self._win.imageWidget._inputType + " / "
                                      + COLOR_MODES.get(self._win.imageWidget.color_mode, "Unknown")
                                      + ("" if self._win.imageWidget._isInputValid else " (Invalid)"))
        if self._win.imageWidget._isInputValid:
            self._win.tbValidInput.setStyleSheet(self._inputTypeDefaultStyle)
        else:
            self._win.tbValidInput.setStyleSheet('background-color : red;')

        # Update runtime
        self._win.runtime.setText(str(datetime.timedelta(seconds=round(self.runtime))))

        # Max / min pixels
        isZoomedImage = self._win.imageWidget.is_zoomed()
        xOffset, yOffset, width, height = self._win.imageWidget.get_zoom_region()
        if self._win.imageWidget.color_mode == COLOR_MODE_MONO:
            self._win.maxPixelChannel.hide()
            self._win.minPixelChannel.hide()
            max_val = self._win.imageWidget._max
            min_val = self._win.imageWidget._min
            max_val_roi = self._win.imageWidget._maxRoi if isZoomedImage else None
            min_val_roi = self._win.imageWidget._minRoi if isZoomedImage else None
        else:
            self._win.maxPixelChannel.show()
            self._win.minPixelChannel.show()
            max_val = self._win.imageWidget._max[self._win.maxPixelChannel.currentIndex()]
            min_val = self._win.imageWidget._min[self._win.minPixelChannel.currentIndex()]
            max_val_roi = self._win.imageWidget._maxRoi[self._win.maxPixelChannel.currentIndex()] if isZoomedImage else None
            min_val_roi = self._win.imageWidget._minRoi[self._win.minPixelChannel.currentIndex()] if isZoomedImage else None

        self.statistics_update(self._win.maxPixel, max_val, max_val_roi)
        self.statistics_update(self._win.minPixel, min_val, min_val_roi)

        # Update frames information
        self.statistics_update(self._win.frameRateCurrAvg,
                        (self.fps_current, self.fps_average),
                        fmt='%.0f / %.2f',
                        lolimit=self._win.imageWidget._pref['FPSLimit'])
        self.statistics_update(self._win.receiveRateCurrAvg,
                        (self.fps_current_received, self.fps_average_received),
                        fmt='%.0f / %.2f',
                        lolimit=self._win.imageWidget._pref['FPSLimit'])
        self.statistics_update(self._win.nFrames,
                        (self._win.imageWidget.frames_displayed, self._win.imageWidget.frames_received),
                        fmt='%d / %d')
        self.statistics_update(self._win.nMissedFramesCurrAvg,
                            (self.frames_missed_current, self.frames_missed_average),
                            fmt = '%.0f / %.2f')
        self.statistics_update(self._win.nMissedFrames, self._win.imageWidget.frames_missed, fmt='%d')
        self.statistics_update(self._win.nQueueSize,
                            (self._win.imageWidget.get_display_queue_size(), self._win.imageWidget.get_display_max_queue_size()),
                            fmt = '%d / %d')

        # No. of dead pixels
        if isZoomedImage:
            values = (max(self._win.imageWidget.dead_pixels), max(self._win.imageWidget.dead_pixels_roi))
            fmt = '%.0f (%.0f)'
        else:
            values = (max(self._win.imageWidget.dead_pixels))
            fmt = '%.0f'
        self.statistics_update(self._win.deadPixel,
                                values,
                                fmt=fmt,
                                hilimit=self._win.imageWidget._pref["DPXLimit"])

        # Machine resources
        self.statistics_update(self._win.cpuUsage,
                                self.cpu_usage,
                                hilimit=self._win.imageWidget._pref['CPULimit'] if self._win.imageWidget._pref['EnableCPULimit'] else None,
                                callback=self._win.imageWidget._pref['EnableCPULimit'])
        self.statistics_update(self._win.netLoad,
                                self.network_usage,
                                fmt='%.2f',
                                hilimit=self._win.imageWidget._pref['NetLimit'] if self._win.imageWidget._pref['EnableNetLimit'] else None,
                                callback=self._win.imageWidget._pref['EnableNetLimit'])
        self.statistics_update(self._win.netReceived,
                                self._win.imageWidget.MB_received,
                                fmt='%.2f')


        # Update image and zoom section
        self._win.lblXsize.setText(' '.join([str(self._win.imageWidget.x),
                '' if not isZoomedImage else f"({width}|{xOffset}-{xOffset+width})"
                                            ]))
        self._win.lblYsize.setText(' '.join([str(self._win.imageWidget.y),
                '' if not isZoomedImage else f"({height}|{yOffset}-{yOffset+height})"
                                            ]))
        zoomMsg = ""
        zoomMsgStyle = "color: black"
        if (isZoomedImage and width == self._win.imageWidget.ZOOM_LENGTH_MIN
                          and height == self._win.imageWidget.ZOOM_LENGTH_MIN):
            zoomMsg = "Maximum zoom reached"
            zoomMsgStyle = "color: red"
        elif isZoomedImage and width == self._win.imageWidget.ZOOM_LENGTH_MIN:
            zoomMsg = "Maximum zoom in X direction reached"
            zoomMsgStyle = "color: red"
        elif isZoomedImage and height == self._win.imageWidget.ZOOM_LENGTH_MIN:
            zoomMsg = "Maximum zoom in Y direction reached"
            zoomMsgStyle = "color: red"
        self._win.zoomStatusLabel.setText(zoomMsg)
        self._win.zoomStatusLabel.setStyleSheet(zoomMsgStyle)

        # General update TODO: why is this required?
        self._win.setStyleSheet('color: black; font-weight: normal')

        # Update the values on the settings dialog
        self._image_settings_dialog.currentDisplayQueueSize.setText(str(self._win.imageWidget.draw_queue.qsize()))

    def changeimageBlackLimits(self, minVal, maxVal):
        """
        Set the minimum and the maximum for the black settings widgets (slider and the spinner).
        If the value is too big or too small for the slider, factor will be calculated, by which
        the value is multiplied/divided.

        :param minVal: (Number) Minimum possible setting for the slider and the spinner.
        :param maxVal: (Number) Maximum possible setting for the slider and the spinner.
        :return:
        """
        sliderFactor = 1
        if (maxVal > self.SLIDER_MAX_VAL or minVal < self.SLIDER_MIN_VAL):
            sliderFactor = max(self.SLIDER_MIN_VAL / minVal if minVal != 0 else 0,
                               self.SLIDER_MAX_VAL / maxVal if maxVal != 0 else 0,)

        self._imageBlackSliderFactor = sliderFactor
        self._win.imageBlackSlider.setMinimum(int(minVal * sliderFactor))
        self._win.imageBlackSlider.setMaximum(int(maxVal * sliderFactor))

        self._win.imageBlackSpinBox.setMinimum(minVal)
        self._win.imageBlackSpinBox.setMaximum(maxVal)

    def getimageBlackLimits(self):
        """
        Get the current settings for the black limits.

        :return: (Number, Number) Tuple of min and max values.
        """
        return (self._win.imageBlackSpinBox.minimum(),
                self._win.imageBlackSpinBox.maximum())

    def changeimageWhiteLimits(self, minVal, maxVal):
        """
        Set the minimum and the maximum for the black settings widgets (slider and the spinner).
        If the value is too big or too small for the slider, factor will be calculated, by which
        the value is multiplied/divided.

        :param minVal: (Number) Minimum possible setting for the slider and the spinner.
        :param maxVal: (Number) Maximum possible setting for the slider and the spinner.
        :return:
        """
        sliderFactor = 1
        if (maxVal > self.SLIDER_MAX_VAL or minVal < self.SLIDER_MIN_VAL):
            sliderFactor = max(self.SLIDER_MIN_VAL / minVal if minVal != 0 else 0,
                               self.SLIDER_MAX_VAL / maxVal if maxVal != 0 else 0,)

        self._imageWhiteSliderFactor = sliderFactor
        self._win.imageWhiteSlider.setMinimum(int(minVal * sliderFactor))
        self._win.imageWhiteSlider.setMaximum(int(maxVal * sliderFactor))

        self._win.imageWhiteSpinBox.setMinimum(minVal)
        self._win.imageWhiteSpinBox.setMaximum(maxVal)

    def getimageWhiteLimits(self):
        """
        Get the current settings for the white limits.

        :return: (Number, Number) Tuple of min and max values.
        """
        return (self._win.imageWhiteSpinBox.minimum(),
                self._win.imageWhiteSpinBox.maximum())

    def updateGuiBlack(self, value):
        """
        Update values on the black slider and spinner.

        :param value: (Number) Value to be set to the black slider and spinner.
        :return:
        """
        # Update the text spinner
        self._win.imageBlackSpinBox.blockSignals(True)
        self._win.imageBlackSpinBox.setValue(value)
        self._win.imageBlackSpinBox.blockSignals(False)

        self._win.imageBlackSlider.blockSignals(True)
        self._win.imageBlackSlider.setValue(int(value * self._imageBlackSliderFactor))
        self._win.imageBlackSlider.blockSignals(False)

    def updateGuiWhite(self, value):
        """
        Update values on the white slider and spinner.

        :param value: (Number) Value to be set to the white slider and spinner.
        :return:
        """
        # Update the text spinner
        self._win.imageWhiteSpinBox.blockSignals(True)
        self._win.imageWhiteSpinBox.setValue(value)
        self._win.imageWhiteSpinBox.blockSignals(False)

        # Update the slider
        self._win.imageWhiteSlider.blockSignals(True)
        self._win.imageWhiteSlider.setValue(int(value * self._imageWhiteSliderFactor))
        self._win.imageWhiteSlider.blockSignals(False)



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""

import sys
import os.path
import psutil
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.Qt import uic
from pyqtgraph.parametertree import Parameter
from .view import Configure
from .model import DataSource as DataReceiver
from .control import ScopeController

form_path = os.path.join(os.path.dirname(__file__), "ui/scope.ui")
form_class = uic.loadUiType(form_path)[0]


class ScopeWindow(QtWidgets.QMainWindow, form_class):
    def __init__(self, parent=None):
        super(ScopeWindow, self).__init__(parent=parent)
        self._proc = psutil.Process()
        self.setupUi(self)
        self.show()


warning_path = os.path.join(os.path.dirname(__file__), "ui/warning.ui")
warning_class = uic.loadUiType(warning_path)[0]


class WarningDialog(QtWidgets.QDialog, warning_class):
    def __init__(self, parent=None):
        super(WarningDialog, self).__init__(parent=parent)
        self.setupUi(self)


def scope(cfg, **kwargs):
    """
    Main function for scope display

    :return:
    """

    # Check for an instance of a QtWidgets.QApplication, if so use it...
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
    else:
        print('QApplication instance already exists: %s' % str(app))

    w = ScopeWindow()

    configure = Configure(cfg, **kwargs)
    parameters = Parameter.create(
        name="params", type="group", children=configure.parse())
    w.parameterPane.setParameters(parameters, showTop=False)
    pvmap = configure.pvs

    warning = WarningDialog(None)

    default_pv = list(pvmap.values())[0] if pvmap else None    
    model = DataReceiver(default=default_pv)
    controller = ScopeController(w, model, parameters, WARNING=warning)
    controller.default_config(**kwargs)

    parameters.sigTreeStateChanged.connect(controller.parameter_change)

    w.show()
    sys.exit(app.exec_())



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import unittest
from configparser import ConfigParser
from c2dataviewer.control.striptool_config import StriptoolConfig
from c2dataviewer.control.pvconfig import PvConfig
from c2dataviewer.view import StripToolConfigure


class TestStriptoolConfig(unittest.TestCase):
    def test_config(self):
        raw = """
[STRIPTOOL]
DefaultProtocol = ca
Chan1.PV = Foo:Bar
Chan2.PV = pva://Bar:Baz
Chan1.Color = #000000
Chan2.Color = #0000FF
"""
        parser = ConfigParser()
        parser.read_string(raw)
        cfg = StriptoolConfig(parser)
        expected = {
            'Foo:Bar': PvConfig('Foo:Bar', '#000000', 'ca'),
            'Bar:Baz': PvConfig('Bar:Baz', '#0000FF', 'pva')
        }

        self.assertEqual(len(cfg.pvs), len(expected))

        for pv in cfg.pvs.values():
            self.assertTrue(pv.pvname in expected)
            e = expected[pv.pvname]
            self.assertEqual(e.pvname, pv.pvname)
            self.assertEqual(e.color, pv.color)
            self.assertEqual(e.proto, pv.proto)
            del expected[pv.pvname]

    def get_display_val(self, data, val):
        for child in data['children']:
            if child['name'] == val:
                ret_val = child['value']
                return ret_val

    def test_autoscale(self):
        #Does autoscale setting in app specific section take precedence
        raw1 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        AUTOSCALE=True

        [DISPLAY]
        AUTOSCALE=False
        AVERAGE=1
        HISTOGRAM=False
        N_BIN=100
        REFRESH=100
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)

        self.assertTrue(self.get_display_val(data=display, val='Autoscale'))

        #When autoscale setting absent in app specific section, but present in DISPLAY
        raw2 = """
        [STRIPTOOL]
        DefaultProtocol = ca

        [DISPLAY]
        AUTOSCALE=False
        AVERAGE=1
        HISTOGRAM=False
        N_BIN=100
        REFRESH=100
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        
        self.assertFalse(self.get_display_val(data=display, val='Autoscale'))

        #When autoscale setting absent in both app specific and in DISPLAY sections,
        #is default selected
        raw3 = """
        [STRIPTOOL]
        DefaultProtocol = ca

        [DISPLAY]
        AVERAGE=1
        HISTOGRAM=False
        N_BIN=100
        REFRESH=100
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)

        self.assertTrue(self.get_display_val(data=display, val='Autoscale'))



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import numpy as np
import pyqtgraph
import pvaccess as pva
from ..model import ConnectionState
import math

class ScopeControllerBase:
    def __init__(self, widget, model, parameters, warning, channels=[], **kwargs):
        self._win = widget
        self.model = model
        self.parameters = parameters
        self.data = None
        # refresh frequency: every 100 ms by default
        self.refresh = 100

        self.plotting_started = False

        self.timer = pyqtgraph.QtCore.QTimer()
        self.timer.timeout.connect(self._win.graphicsWidget.update_drawing)
        self._win.graphicsWidget.set_autoscale(parameters.child('Display', 'Autoscale').value())
        self._win.graphicsWidget.set_model(self.model)

        self._warning = warning
        self._warning.warningConfirmButton.clicked.connect(lambda: self.accept_warning())

        self.arrays = np.array([])
        self.lastArrays = 0

        self.default_trigger = None
        self.trigger_is_monitor = False
        self.trigger_auto_scale = False

        self._win.graphicsWidget.set_histogram(parameters.child('Display', 'Histogram').value())
        single_axis = parameters.child('Display', 'Single axis').value()
        self._win.graphicsWidget.setup_plot(channels=channels, single_axis=single_axis)

        display_mode = parameters.child('Display', 'Mode').value()
        if(display_mode != "normal"):
            self.set_display_mode(display_mode)
            
        # timer to update status with statistics data
        self.status_timer = pyqtgraph.QtCore.QTimer()
        self.status_timer.timeout.connect(self.update_status)
        self.status_timer.start(1000)

    def set_display_mode(self, val):
        self._win.graphicsWidget.set_display_mode(val)
        self._win.graphicsWidget.setup_plot()
        # Disable multiaxis for the FFT and PSD modes. As at the time of the
        # writing this code pyqtgraph does not support logarithmic scale in multiaxis configuration.
        if self._win.graphicsWidget.fft or self._win.graphicsWidget.psd:
            self.parameters.child("Display").child("Single axis").setReadonly()
        else:
            self.parameters.child("Display").child("Single axis").setWritable()

    def set_trigger_pv(self, pvname):
        if self._win.graphicsWidget.plotting_started:
            self.notify_warning("Stop plotting first before changing trigger PV")
            return
                
        pvname = pvname.strip()
        if pvname == '':
            return
        
        proto = 'ca'
        name = pvname
        
        if "://" in name:
            proto, name = pvname.split("://")

        try:
            pvfields = self.model.update_trigger(name, proto=proto.lower())
            child = self.parameters.child("Trigger").child("Time Field")
            if proto == 'ca':
                child.hide()
                self._win.graphicsWidget.trigger.trigger_time_field = 'timeStamp'
            else:
                self._win.graphicsWidget.trigger.trigger_time_field = None
                pvfields.insert(0, 'None')
                child.show()
                child.setLimits(pvfields)
                child.setValue('None')                                
                                
        except Exception as e:
            self.notify_warning("Channel {}://{} timed out. \n{}".format(proto, name, repr(e)))


    def default_config(self, **kwargs):
        buffer_unit = kwargs.get('buffer_unit', 'Samples')
            
        max_length = self.parameters.child("Acquisition").child("Buffer (%s)" % buffer_unit).value()
        if max_length and buffer_unit == 'Samples':
            self.update_buffer_samples(int(max_length))
            
        self._win.graphicsWidget.set_binning(self.parameters.child("Display").child("Num Bins").value())

        refresh = self.parameters.child("Display").child("Refresh").value()
        if refresh:
            self.set_freshrate(refresh)
            
        self.default_trigger = kwargs.get("trigger", None)
        if self.default_trigger is not None:
            self.set_trigger_pv(self.default_trigger)

        try:
            self.trigger_auto_scale = self.parameters.child("Trigger").child("Autoscale Buffer").value()
        except:
            pass

        
    def update_buffer_samples(self, size):
        """
        Sets number of samples in buffer
        
        :param size  size of buffer in number of samples
        """
        self._win.graphicsWidget.update_buffer(size)
        self.parameters.child("Acquisition").child("Buffer (Samples)").setValue(self._win.graphicsWidget.max_length)
            
    def parameter_change(self, params, changes):
        """
        
        :param params:
        :param changes:
        :return:
        """
        for param, change, data in changes:
            if change == "value":
                path = self.parameters.childPath(param)
                if path is not None:
                    childName = '.'.join(path)
                else:
                    childName = param.name()

                if childName == "Trigger.PV":
                    if data is None:
                        return
                    self.set_trigger_pv(data)
                elif childName == "Trigger.Mode":
                    self.set_trigger_mode(data)
                elif childName == "Trigger.Threshold":
                    self._win.graphicsWidget.trigger.trigger_level = data
                elif childName == "Trigger.Data Time Field":
                    self._win.graphicsWidget.trigger.data_time_field = data
                elif childName == "Trigger.Time Field":
                    self._win.graphicsWidget.trigger.trigger_time_field = data
                elif childName == "Trigger.Autoscale Buffer":
                    self.trigger_auto_scale = data
                elif childName == "Acquisition.Buffer (Samples)":
                    self._win.graphicsWidget.update_buffer(data)
                elif childName == "Acquisition.Freeze":
                    self._win.graphicsWidget.is_freeze = data
                elif childName == "Display.Mode":
                    self.set_display_mode(data)
                elif childName == "Display.FFT filter":
                    self._win.graphicsWidget.set_fft_filter(data)
                elif childName == "Display.Exp moving avg":
                    self._win.graphicsWidget.set_average(data)
                elif childName == "Display.Autoscale":
                    self._win.graphicsWidget.set_autoscale(data)
                elif childName == "Display.Single axis":
                    self._win.graphicsWidget.setup_plot(single_axis=data)
                elif childName == "Display.Histogram":
                    self._win.graphicsWidget.set_histogram(data)
                elif childName == "Display.Num Bins":
                    self._win.graphicsWidget.set_binning(data)
                elif childName == "Display.Refresh":
                    self.set_freshrate(data)

    def set_freshrate(self, value):
        """
        Set time to refresh
            
        :param value: time interval to plot, in second
        :return:
        """
        plotting_started = self._win.graphicsWidget.plotting_started
        self.stop_plotting()
        self.refresh = value*1000.0
        if plotting_started:
            self.start_plotting()


    def start_plotting(self):
        """
            
        :return:
        """

        try:
            self.timer.timeout.disconnect()
            self.timer.stop()
        except Exception:
            pass

        self.stop_trigger()
        
        # Setup free run plotting
        if not self._win.graphicsWidget.trigger_mode():
            self.timer.timeout.connect(self._win.graphicsWidget.update_drawing)
            self.timer.start(int(self.refresh))
        else:
            self.start_trigger()

        self._win.graphicsWidget.notify_plotting_started(True)
        
    def stop_plotting(self):
        """

        :return:
        """
        self._win.graphicsWidget.notify_plotting_started(False)

        self.timer.stop()
        self.stop_trigger()
                    
    def set_trigger_mode(self, value):
        """
        Set trigger mode.
    
        :param value:
        :return:
        """
        trigger_mode = value != 'none'
        if trigger_mode != self._win.graphicsWidget.trigger_mode():
            if self._win.graphicsWidget.plotting_started:
                action = 'on' if trigger_mode else 'off'
                self.notify_warning('Stop plotting first before turning trigger %s' % (action))
                return
            
        self._win.graphicsWidget.set_trigger_mode(trigger_mode)
            
        trigger_type = None
        if trigger_mode:
            self._win.graphicsWidget.trigger.set_trigger_type(value)
            
    def start_trigger(self):
        """
        Process to start DAQ in trigger mode
            
        :return:
        """
        if not self.trigger_is_monitor:
            if self.model.trigger is None:
                raise Exception('Trigger PV is not set')
            
            if self._win.graphicsWidget.trigger.data_time_field is None:
                raise Exception('Data time field is not set')
            
            if self._win.graphicsWidget.trigger.trigger_time_field is None:
                raise Exception('Trigger time field is not set')
            
            self.model.start_trigger(self._win.graphicsWidget.trigger.data_callback)
            self.trigger_is_monitor = True

    def stop_trigger(self):
        """
        Stop trigger mode
        :return:
        """
        self.trigger_is_monitor = False
        self.model.stop_trigger()
        
    def accept_warning(self):
        """
            
        :return:
        """
        self._warning.close()

    def notify_warning(self, msg):
        #close previous warning
        self.accept_warning()
            
        self._warning.warningTextBrowse.setText(msg)
        self._warning.show()

        
    def update_status(self):
        """
        Update statistics status.
            
        :return:
        """
        # Update display
        single_axis_child = self.parameters.child("Display", "Single axis")
        if single_axis_child.value() != self._win.graphicsWidget.single_axis:
            single_axis_child.setValue(self._win.graphicsWidget.single_axis)

        # Update statistics
        with self._win._proc.oneshot():
            cpu = self._win._proc.cpu_percent(None)

        # TODO algorithm to calculate Array/sec
        arraysReceived = self._win.graphicsWidget.arraysReceived
        n = arraysReceived - self.lastArrays
        self.lastArrays = arraysReceived
        self.arrays = np.append(self.arrays, n)[-10:]

        for q in self.parameters.child("Statistics").children():
            if q.name() == 'CPU':
                q.setValue(cpu)
            elif q.name() == 'Lost Arrays':
                q.setValue(self._win.graphicsWidget.lostArrays)
            elif q.name() == 'Tot. Arrays':
                q.setValue(self._win.graphicsWidget.arraysReceived)
            elif q.name() == 'Arrays/Sec':
                q.setValue(self.arrays.mean())
            elif q.name() == 'Bytes/Sec':
                q.setValue(self.arrays.mean() * self._win.graphicsWidget.data_size)
            elif q.name() == 'Rate':
                q.setValue(self._win.graphicsWidget.fps)

        try:
            for q in self.parameters.child("Trigger").children():
                if q.name() == "Trig Status":
                    stat_str = 'Disconnected'
                    if self.trigger_is_monitor:
                        stat_str = self._win.graphicsWidget.trigger.status()
                    q.setValue(stat_str)
                elif q.name() == "Trig Value":
                    q.setValue(str(self._win.graphicsWidget.trigger.trigger_value))
        except:
            pass

        #handle any auto-adjustments
        if self.trigger_auto_scale and self._win.graphicsWidget.trigger_mode():
            if self._win.graphicsWidget.trigger.missed_triggers > 0:
                newsize = self._win.graphicsWidget.trigger.missed_adjust_buffer_size

                # Round up to 3 10's places
                # ex. 12345 would be rounded to 12400
                precision = 3
                exp = math.ceil(math.log10(newsize))
                roundunit = 10**max(exp - precision, 0)
                newsize = math.ceil(newsize / roundunit) * roundunit

                self.update_buffer_samples(newsize)
            
        



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

class ScopeConfigureBase:
    def __init__(self, params, **kwargs):
        self.params = params
        self.default_trigger = kwargs.get("trigger", None)

    def add_source_aquisition_props(self, children, section):
        """
        Add acquisition information specific to the application data source.  This
        should be implemented by the child class
        """
        return children

    def assemble_acquisition(self, section=None, buffer_unit='Samples'):
        """
        Assemble acquisition information

        :param section:
        :return:
        """

        children = []

        buffer_size = None
        try:
            buffer_size = section["BUFFER"]
        except:
            pass
        
        children += [
            {"name": "Freeze", "type": "bool", "value": False},
            {"name": "Buffer (%s)" % buffer_unit, "type": "int", "value": buffer_size, "siPrefix": False, 'decimals': 20}
        ]

        acquisition = {"name": "Acquisition", "type":"group", "children": children}
        
        return acquisition

    
    def assemble_display(self, section=None, app_section_key=None, default_autoscale=None):
        """
        Assemble display information

        :param section: DISPLAY section in config file
        :param app_section_key: type of app
        :param default_autoscale: default autoscale values for each app
        :return:
        """
        # If AUTOSCALE set in the app specific sections in the config file
        autoscale = self.params.get(app_section_key, "AUTOSCALE", fallback=None)

        if autoscale is None:
            # Else if AUTOSCALE set in the DISPLAY specific section in the config file
            autoscale = self.params.get("DISPLAY", "AUTOSCALE", fallback=None)
        
        # If AUTOSCALE not set anywhere in the config file, set default value
        if autoscale is None:
            autoscale = default_autoscale
        elif str(autoscale).upper().strip() in ["TRUE"]:
            autoscale = True
        else:
            autoscale = False

        if section is None:
            display = {"name": "Display", "type": "group",
                       "expanded": True,
                       "children": [
                {"name": "Mode", "type": "list", "limits": {
                    "Normal": "normal",
                    "FFT": "fft",
                    "PSD": "psd",
                    "Diff": "diff",
                    "X vs Y": "xy",
                }, "value": "Normal"},
                {"name": "FFT filter", "type": "list", "limits": {
                    "None" : "none",
                    "Hamming" : "hamming"
                }, "value": "None"},
                {"name": "Exp moving avg", "type": "int", "value": 1, "limits" : (1, 1e+10)},
                {"name": "Autoscale", "type": "bool", "value": autoscale},
                {"name": "Single axis", "type": "bool", "value": True},
                {"name": "Histogram", "type": "bool", "value": False},
                {"name": "Num Bins", "type": "int", "value": 100},
                {"name": "Refresh", "type": "float", "value": 0.1, "siPrefix": True, "suffix": "s"},
            ]}
        else:
            display_mode = section.get("MODE", "normal").lower().strip()
            if display_mode not in ["normal", "fft", "psd", "diff", "xy"]:
                display_mode = "normal"

            fft_filter = section.get("FFT_FILTER", None)
            if fft_filter is not None:
                if fft_filter.lower().strip() in ["none", "hamming"]:
                    fft_filter = fft_filter.lower().strip().capitalize()
                else:
                    # Disable filter if config not valid
                    fft_filter = "none"

            try:
                n_average = section.getint("AVERAGE", 1)
                n_average = n_average if n_average > 0 else 1
            except ValueError:
                n_average = 1

            single_axis = section.get("SINGLE_AXIS", None)
            if single_axis is not None:
                if single_axis.upper().strip() in ["FALSE"]:
                    single_axis = False
                else:
                    single_axis = True
            else:
                single_axis = True

            histogram = section.get("HISTOGRAM", None)
            if histogram is not None:
                if histogram.upper().strip() in ["TRUE"]:
                    histogram = True
                else:
                    # set auto scale value to None if a "None" string comes from configuration
                    histogram = False
            else:
                histogram = False

            try:
                n_binning = section.getint("N_BIN", 100)
            except ValueError:
                n_binning = 1

            try:
                refresh = section.getfloat("REFRESH", 100)
            except ValueError:
                refresh = 1

            display = {"name": "Display", "type": "group", "expanded": True,
                       "children": [
                {"name": "Mode", "type": "list", "limits": {
                    "Normal": "normal",
                    "FFT": "fft",
                    "PSD": "psd",
                    "Diff": "diff",
                    "X vs Y": "xy",
                }, "value": display_mode},
                {"name": "FFT filter", "type": "list", "limits": {
                    "None" : "none",
                    "Hamming" : "hamming"
                }, "value": fft_filter},
                {"name": "Exp moving avg", "type": "int", "value": n_average, "limits" : (1, 1e+10)},
                {"name": "Autoscale", "type": "bool", "value": autoscale},
                {"name": "Single axis", "type": "bool", "value": single_axis},
                {"name": "Histogram", "type": "bool", "value": histogram},
                {"name": "Num Bins", "type": "int", "value": n_binning},
                {"name": "Refresh", "type": "float", "value": refresh / 1000., "siPrefix": True, "suffix": "s"},
            ]}

        return display


    def assemble_trigger(self, section=None):
        trigger_pv = self.default_trigger
        trigger_mode = 'none'

        if section:
            # Trigger PV name and protocol, which comes with format of proto://pv_name
            # the protocol is either ca:// for channel access or pva:// for pvAccess
            trigger_pv = section.get("TRIGGER", None)
            if self.default_trigger is not None:
                trigger_pv = self.default_trigger

            if trigger_pv is not None and trigger_pv.upper().strip() == "NONE":
                # set trigger PV value to None if a "None" string comes from configuration
                trigger_pv = None
            if trigger_pv is None:
                trigger_mode = None
            else:
                trigger_mode = section.get("TRIGGER_MODE", None)
            if trigger_mode is not None:
                trigger_mode = trigger_mode.lower().strip()
                if trigger_mode == 'off':
                    trigger_mode = 'none'
            else:
                trigger_mode = 'none'
        
        cfg ={"name": "Trigger",
              "type": "group",
              "expanded": False,
              "children" : [
                  { "name" :  "Mode", "type": "list", "limits": {
                    "Off" : "none",
                    "On change" : "onchange",
                    "Greater than threshold" : "gtthreshold",
                    "Lesser than threshold" : "ltthreshold"
                    },
                  "value" : trigger_mode
                 },
                  {"name": "PV", "type": "str", "value": trigger_pv},
                  {"name": "Trig Status", "type": "str", "value": "", "readonly": True},
                  {"name": "Trig Value", "type": "str", "value": "", "readonly": True},
                  {"name": "Time Field", "type": "list", "limits" :
                   [ "None" ], "default" : "None", "visible" : False},
                  {"name": "Data Time Field", "type": "list", "limits" :["None"], "default" : "None" },
                  {"name": "Autoscale Buffer", "type": "bool", "value" : True},
                  {"name": "Threshold", "type": "float", "value": 0.0},
              ]}
        return cfg
    
    def assemble_statistics(self):
        statistics = {"name": "Statistics", "type": "group",
                      "expanded":False,
                      "children": [
            {"name": "CPU", "type": "float", "value": 0, "readonly": True, "suffix": "%"},
            {"name": "Lost Arrays", "type": "int", "value": 0, "readonly": True},
            {"name": "Tot. Arrays", "type": "int", "value": 0, "readonly": True, "siPrefix": True},
            {"name": "Arrays/Sec", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "/sec"},
            {"name": "Bytes/Sec", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "/sec"},
            {"name": "Rate", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "Frames/sec"},
        ]}
        return statistics
